import asyncio
from pathlib import Path
import shutil
import tempfile
from typing import AsyncGenerator, List, Sequence, Optional
import re
from typing import Any, Mapping
import uuid
from loguru import logger
from datetime import datetime
from pydantic import Field
from autogen_core import CancellationToken, ComponentModel, Component
from autogen_core.models import (
    ChatCompletionClient,
    UserMessage,
    SystemMessage,
)
from pydantic import BaseModel
from typing_extensions import Self

from autogen_agentchat.agents import BaseChatAgent
from autogen_core.code_executor import CodeBlock, CodeExecutor
from autogen_core.model_context import (
    ChatCompletionContext,
    TokenLimitedChatCompletionContext,
)
from autogen_agentchat.base import Response
from autogen_agentchat.state import BaseState
from autogen_agentchat.messages import (
    BaseAgentEvent,
    BaseChatMessage,
    TextMessage,
    MessageFactory,
)
from autogen_core.code_executor import CodeResult
from autogen_ext.code_executors.local import LocalCommandLineCodeExecutor
from autogen_ext.code_executors.docker import DockerCommandLineCodeExecutor

from ....ui_backend.utils import thread_to_context
from ._utils import exec_command_umask_patched

from ..approval_guard import BaseApprovalGuard
from ..guarded_action import ApprovalDeniedError, TrivialGuardedAction

DockerCommandLineCodeExecutor._execute_command = exec_command_umask_patched  # type: ignore


def _extract_markdown_code_blocks(markdown_text: str) -> List[CodeBlock]:
    pattern = re.compile(r"```(?:\s*([\w\+\-]+))?\n([\s\S]*?)```")
    matches = pattern.findall(markdown_text)
    code_blocks: List[CodeBlock] = []
    for match in matches:
        language = match[0].strip() if match[0] else ""
        code_content = match[1]
        code_blocks.append(CodeBlock(code=code_content, language=language))
    return code_blocks


async def _invoke_action_guard(
    thread: Sequence[BaseChatMessage | BaseAgentEvent],
    delta: Sequence[BaseChatMessage | BaseAgentEvent],
    code_message: TextMessage,
    agent_name: str,
    model_client: ChatCompletionClient,
    approval_guard: BaseApprovalGuard | None,
) -> None:
    # Get approval for the coding request. We could conceivably do extra work to enable interactive approval here,
    # but the value for many users is likely to be low, as it may not be appropriate to assume knowledge of coding,
    # and thus the user will not have the context necessary to approve/deny the incremental execution of code blocks.
    guarded_action = TrivialGuardedAction("coding", baseline_override="maybe")

    # Note that delta already contains the code message.
    assert delta[-1] == code_message

    thread = list(thread) + list(delta)

    context = thread_to_context(
        thread,
        agent_name,
        is_multimodal=model_client.model_info["vision"],
    )
    action_description_for_user = TextMessage(
        content="Do you want to execute the code above?",
        source=agent_name,
    )

    await guarded_action.invoke_with_approval(
        {}, code_message, context, approval_guard, action_description_for_user
    )


async def _coding_and_debug(
    system_prompt: str,
    thread: Sequence[BaseChatMessage],
    agent_name: str,
    model_client: ChatCompletionClient,
    code_executor: CodeExecutor,
    max_debug_rounds: int,
    cancellation_token: CancellationToken,
    model_context: ChatCompletionContext,
    approval_guard: BaseApprovalGuard | None,
) -> AsyncGenerator[TextMessage | bool, None]:
    """Write and debug code using the model and executor.

    It generates code based on the system prompt and the thread of messages,
    executes it, and returns the output. It continues to generate and execute
    code until the maximum number of debug rounds is reached or the code
    executes successfully.

    When the cancellation token is set, the execution will stop.
    Args:
        system_prompt (str): The system prompt to guide the model.
        thread (Sequence[BaseChatMessage]): The thread of messages to use as context.
        agent_name (str): The name of the agent.
        model_client (ChatCompletionClient): The model client to use for code generation.
        code_executor (CodeExecutor): The code executor to use for executing the code.
        max_debug_rounds (int): The maximum number of debug rounds to perform.
        cancellation_token (CancellationToken): The cancellation token to stop execution.
        model_context (ChatCompletionContext): The context for the model.
        approval_guard (ApprovalGuard | None): The approval guard to use for code execution.

    Yields:
        TextMessage: The intermediate messages generated by the model and executor.
        bool: A flag indicating whether any code execution was performed.

    Raises:
        ApprovalDeniedError: If the user denies the approval for the coding request.
    """
    # The list of new messages to be added to the thread.
    delta: Sequence[BaseChatMessage | BaseAgentEvent] = []
    executed_code = False

    for i in range(max_debug_rounds):
        # Add system prompt as the last message before generation
        current_thread = (
            list(thread)
            + list(delta)
            + [TextMessage(source="user", content=system_prompt)]
        )

        # create an LLM context from system message, global chat history, and inner messages
        context = [SystemMessage(content=system_prompt)] + thread_to_context(
            current_thread,
            agent_name,
            is_multimodal=model_client.model_info["vision"],
        )

        # Re-initialize model context to meet token limit quota
        try:
            await model_context.clear()
            for msg in context:
                await model_context.add_message(msg)
            token_limited_context = await model_context.get_messages()
        except Exception:
            token_limited_context = context
        # Generate code using the model.
        create_result = await model_client.create(
            messages=token_limited_context, cancellation_token=cancellation_token
        )
        assert isinstance(create_result.content, str)
        code_msg = TextMessage(
            source=agent_name + "-llm",
            metadata={"internal": "no", "type": "potential_code"},
            content=create_result.content,
        )
        # add LLM's response to the current thread.
        delta.append(code_msg)
        yield code_msg

        # extract code blocks from the LLM's response
        code_block_list = _extract_markdown_code_blocks(create_result.content)
        # if no code to execute, return
        if len(code_block_list) == 0:
            break

        # Now that we know we have code to execute, make sure we have permission to do so.
        # Note that we will be getting this permission per round of code/execute/debug.
        if approval_guard is not None:
            await _invoke_action_guard(
                thread=thread,
                delta=delta,
                code_message=code_msg,  # note that this `code_msg is delta[-1]`
                agent_name=agent_name,
                model_client=model_client,
                approval_guard=approval_guard,
            )

        code_output_list: List[str] = []
        exit_code_list: List[int] = []
        executed_code = True
        try:
            for cb in code_block_list:
                # execute the code block
                exit_code: int = 1
                encountered_exception: bool = False
                code_output: str = ""
                result: CodeResult | None = None
                try:
                    result = await code_executor.execute_code_blocks(
                        [cb], cancellation_token
                    )
                    exit_code = result.exit_code or 0
                    code_output = result.output
                except Exception as e:
                    code_output = str(e)
                    encountered_exception = True
                if encountered_exception or result is None:
                    code_output = f"An exception occurred while executing the code block: {code_output}"
                elif code_output.strip() == "":
                    # No output
                    code_output = f"The script ran but produced no output to console. The POSIX exit code was: {result.exit_code}. If you were expecting output, consider revising the script to ensure content is printed to stdout."
                elif exit_code != 0:
                    # Error
                    code_output = f"The script ran, then exited with an error (POSIX exit code: {result.exit_code})\nIts output was:\n{result.output}"
                code_output_list.append(code_output)
                code_output_msg = TextMessage(
                    source=agent_name + "-executor",
                    metadata={"internal": "no", "type": "code_execution"},
                    content=f"Execution result of code block {i + 1}:\n```console\n{code_output}\n```",
                )
                exit_code_list.append(exit_code)
                yield code_output_msg

            final_code_output = ""
            for i, code_output in enumerate(code_output_list):
                final_code_output += f"\n\nExecution Result of Code Block {i + 1}:\n```console\n{code_output}\n```"

            # add executors response to thread.
            executor_msg = TextMessage(
                source=agent_name + "-executor",
                metadata={"internal": "yes"},
                content=final_code_output,
            )
            delta.append(executor_msg)
            yield executor_msg

            # break if the code execution was successful
            if all([code_output == 0 for code_output in exit_code_list]):
                break
        except asyncio.TimeoutError:
            # If the task times out, we treat it as an error.
            executor_msg = TextMessage(
                source=agent_name + "-executor",
                metadata={"internal": "yes"},
                content="Code execution timed out.",
            )
            delta.append(executor_msg)
            yield executor_msg

    yield executed_code


async def _summarize_coding(
    agent_name: str,
    model_client: ChatCompletionClient,
    thread: Sequence[BaseChatMessage | BaseAgentEvent],
    cancellation_token: CancellationToken,
    model_context: ChatCompletionContext,
) -> TextMessage:
    # Create a summary from the inner messages using an extra LLM call.
    input_messages = (
        [SystemMessage(content="You are an agent that can write and debug code")]
        + thread_to_context(
            list(thread), agent_name, is_multimodal=model_client.model_info["vision"]
        )
        + [
            UserMessage(
                content="""
                The above is a transcript of your previous messages and a request that was given to you in the begining.
                You need to summarize them to answer the request given to you. Generate a summary of everything that happened.
                If there was code that was executed, please copy the final code that was executed without errors.
                Don't mention that this is a summary, just give the summary.""",
                source="user",
            )
        ]
    )

    # Re-initialize model context to meet token limit quota
    try:
        await model_context.clear()
        for msg in input_messages:
            await model_context.add_message(msg)
        token_limited_input_messages = await model_context.get_messages()
    except Exception:
        token_limited_input_messages = input_messages

    summary_result = await model_client.create(
        messages=token_limited_input_messages, cancellation_token=cancellation_token
    )
    assert isinstance(summary_result.content, str)
    return TextMessage(
        source=agent_name,
        metadata={"internal": "yes"},
        content=summary_result.content,
    )


class CoderAgentConfig(BaseModel):
    name: str
    model_client: ComponentModel
    description: str = """
    An agent that can write and execute code to solve tasks or use its language skills to summarize, write, solve math and logic problems.
    It understands images and can use them to help it complete the task.
    It can access files if given the path and manipulate them using python code. Use the coder if you want to manipulate a file or read a csv or excel files.
    """
    max_debug_rounds: int = 3
    summarize_output: bool = False
    # Optionally add code_executor config if needed


class CoderAgentState(BaseState):
    chat_history: List[BaseChatMessage] = Field(default_factory=list[BaseChatMessage])
    type: str = Field(default="CoderAgentState")


class CoderAgent(BaseChatAgent, Component[CoderAgentConfig]):
    """An agent capable of writing, executing, and debugging code.

    The agent uses either a local or Docker-based code executor to run the generated code
    in a controlled environment. It maintains a chat history and can be paused/resumed
    during execution.
    """

    component_type = "agent"
    component_config_schema = CoderAgentConfig
    component_provider_override = "magentic_ui.agents.CoderAgent"

    DEFAULT_DESCRIPTION = """
    An agent that can write and execute code to solve tasks or use its language skills to summarize, write, solve math and logic problems.
    It understands images and can use them to help it complete the task.
    It can access files if given the path and manipulate them using python code. Use the coder if you want to manipulate a file or read a csv or excel files.
    In a single step when you ask the agent to do something: it can write code, and then immediately execute the code. If there are errors it can debug the code and try again. 
    """

    system_prompt_coder_template = """
    You are helpful assistant.
    In addition to responding with text you can write code and execute code that you generate.
    The date today is: {date_today}

    Rules to follow for Code:
    - Generate py or sh code blocks in the order you'd like your code to be executed.
    - Code block must indicate language type. Do not try to predict the answer of execution. Code blocks will be automatically executed for you.
    - If you want to stop executing code, make sure to not write any code in your message and your turn will be over.
    - Do not generate code that relies on API keys that you don't have access to. Try different approaches.

    Tips:
    - You don't have to generate code if the task is not related to code, for instance writing a poem, paraphrasing a text, etc.
    - If you are asked to solve math or logical problems, first try to answer them without code and then if needed try to use python to solve them.
    - You have access to the standard Python libraries in addition to numpy, pandas, scikit-learn, matplotlib, pillow, requests, beautifulsoup4.
    - If you need to use an external library, write first a shell script that installs the library first using pip install, then add code blocks to use the library.
    - Always use print statements to output your work and partial results.
    - For showing plots or other visualizations that are not just text, make sure to save them to file with the right extension for them to be displayed.

   VERY IMPORTANT: If you intend to write code to be executed, do not end your response without a code block. If you want to write code you must provide a code block in the current generation.
    """

    def __init__(
        self,
        name: str,
        model_client: ChatCompletionClient,
        model_context_token_limit: int = 128000,
        description: str = DEFAULT_DESCRIPTION,
        max_debug_rounds: int = 3,
        summarize_output: bool = False,
        code_executor: Optional[CodeExecutor] = None,
        work_dir: Path | str | None = None,
        bind_dir: Path | str | None = None,
        use_local_executor: bool = False,
        approval_guard: BaseApprovalGuard | None = None,
    ) -> None:
        """Initialize the CoderAgent.

        Args:
            name (str): The name of the agent
            model_client (ChatCompletionClient): The language model client to use
            description (str, optional): Description of the agent's capabilities. Default: DEFAULT_DESCRIPTION.
            max_debug_rounds (int, optional): Maximum number of code debugging iterations. Default: 3.
            summarize_output (bool, optional): Whether to summarize code execution results. Default: False.
            code_executor (Optional[CodeExecutor], optional): Custom code executor to use. Default: None.
            work_dir (Path | str | None, optional): Working directory for code execution. Default: None.
            bind_dir (Path | str | None, optional): Directory to bind for Docker executor. Default: None.
            use_local_executor (bool, optional): Whether to use local instead of Docker executor. Default: False.
        """
        super().__init__(name, description)
        self._model_client = model_client
        self._model_context = TokenLimitedChatCompletionContext(
            model_client, token_limit=model_context_token_limit
        )
        self._chat_history: List[BaseChatMessage] = []
        self._max_debug_rounds = max_debug_rounds
        self._summarize_output = summarize_output
        self.is_paused = False
        self._paused = asyncio.Event()
        self._approval_guard = approval_guard

        if work_dir is None:
            self._work_dir = Path(tempfile.mkdtemp())
            self._cleanup_work_dir = True
        else:
            self._work_dir = Path(work_dir)
            self._cleanup_work_dir = False
        if code_executor:
            self._code_executor = code_executor
        elif use_local_executor:
            self._code_executor = LocalCommandLineCodeExecutor(work_dir=self._work_dir)
        else:
            name = f"{name}-{uuid.uuid4()}"
            self._code_executor = DockerCommandLineCodeExecutor(
                container_name=name,
                image="magentic-ui-python-env",
                work_dir=self._work_dir,
                bind_dir=bind_dir,
                delete_tmp_files=True,
            )

    async def lazy_init(self) -> None:
        """Initialize the code executor if it has a start method.

        This method is called after initialization to set up any async resources
        needed by the code executor.
        """
        if self._code_executor:
            # check if the code executor has a start method
            if hasattr(self._code_executor, "start"):
                # TODO: we should add a no-op start() method to the base class.
                await self._code_executor.start()  # type: ignore

    async def close(self) -> None:
        """Clean up resources used by the agent.

        This method:
        - Stops the code executor
        - Removes the work directory if it was created
        - Closes the model client
        """
        logger.info("Closing Coder...")
        await self._code_executor.stop()
        # Remove the work directory if it was created.
        if self._cleanup_work_dir and self._work_dir.exists():
            await asyncio.to_thread(shutil.rmtree, self._work_dir)
        # Close the model client.
        await self._model_client.close()

    async def pause(self) -> None:
        """Pause the agent by setting the paused state."""
        self.is_paused = True
        self._paused.set()

    async def resume(self) -> None:
        """Resume the agent by clearing the paused state."""
        self.is_paused = False
        self._paused.clear()

    @property
    def produced_message_types(self) -> Sequence[type[BaseChatMessage]]:
        """Get the types of messages produced by the agent."""
        return (TextMessage,)

    async def on_messages(
        self, messages: Sequence[BaseChatMessage], cancellation_token: CancellationToken
    ) -> Response:
        """Handle incoming messages and return a single response. Calls the on_messages_stream."""
        response: Response | None = None
        async for message in self.on_messages_stream(messages, cancellation_token):
            if isinstance(message, Response):
                response = message
        assert response is not None
        return response

    async def on_messages_stream(
        self, messages: Sequence[BaseChatMessage], cancellation_token: CancellationToken
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage | Response, None]:
        """Handle incoming messages and yield responses as a stream. Append the request to agents chat history."""
        if self.is_paused:
            yield Response(
                chat_message=TextMessage(
                    content="The Coder is paused.",
                    source=self.name,
                    metadata={"internal": "yes"},
                )
            )
            return
        self._chat_history.extend(messages)
        last_message_received: BaseChatMessage = messages[-1]
        inner_messages: List[BaseChatMessage] = []

        # Set up the cancellation token for the code execution.
        code_execution_token = CancellationToken()

        # Cancel the code execution if the handler's cancellation token is set.
        cancellation_token.add_callback(lambda: code_execution_token.cancel())

        # Set up background task to monitor the pause event and cancel the code execution if paused.
        async def monitor_pause() -> None:
            await self._paused.wait()
            code_execution_token.cancel()

        monitor_pause_task = asyncio.create_task(monitor_pause())

        system_prompt_coder = self.system_prompt_coder_template.format(
            date_today=datetime.now().strftime("%Y-%m-%d")
        )

        try:
            executed_code = False
            # Run the code execution and debugging process.
            async for msg in _coding_and_debug(
                system_prompt=system_prompt_coder,
                thread=self._chat_history,
                agent_name=self.name,
                model_client=self._model_client,
                code_executor=self._code_executor,
                max_debug_rounds=self._max_debug_rounds,
                cancellation_token=code_execution_token,
                model_context=self._model_context,
                approval_guard=self._approval_guard,
            ):
                if isinstance(msg, bool):
                    executed_code = msg
                    break
                inner_messages.append(msg)
                self._chat_history.append(msg)
                yield msg

            # New conditional block based on the configuration flag.
            if self._summarize_output and executed_code:
                summary_msg = await _summarize_coding(
                    agent_name=self.name,
                    model_client=self._model_client,
                    thread=[last_message_received] + inner_messages,
                    cancellation_token=code_execution_token,
                    model_context=self._model_context,
                )
                self._chat_history.append(summary_msg)
                yield Response(chat_message=summary_msg, inner_messages=inner_messages)
            else:
                # Instead of only executor output, return a transcript of all code and execution steps.
                combined_output = ""
                for txt_msg in inner_messages:
                    assert isinstance(txt_msg, TextMessage)
                    combined_output += f"{txt_msg.content}\n"
                final_response_msg = TextMessage(
                    source=self.name,
                    metadata={"internal": "yes"},
                    content=combined_output or "No output.",
                )
                # TODO: do we not want to add this to the chat history?
                yield Response(
                    chat_message=final_response_msg, inner_messages=inner_messages
                )
        except ApprovalDeniedError:
            # If the user denies the approval, we respond with a message.
            yield Response(
                chat_message=TextMessage(
                    content="The user did not approve the code execution.",
                    source=self.name,
                    metadata={"internal": "no"},
                ),
                inner_messages=inner_messages,
            )
        except asyncio.CancelledError:
            # If the task is cancelled, we respond with a message.
            yield Response(
                chat_message=TextMessage(
                    content="The task was cancelled by the user.",
                    source=self.name,
                    metadata={"internal": "yes"},
                ),
                inner_messages=inner_messages,
            )
        except Exception as e:
            logger.error(f"Error in CoderAgent: {e}")
            # add to chat history
            self._chat_history.append(
                TextMessage(
                    content=f"An error occurred while executing the code: {e}",
                    source=self.name,
                )
            )
            yield Response(
                chat_message=TextMessage(
                    content="An error occurred while executing the code.",
                    source=self.name,
                    metadata={"internal": "no"},
                ),
                inner_messages=inner_messages,
            )
        finally:
            # Cancel the monitor task.
            try:
                monitor_pause_task.cancel()
                await monitor_pause_task
            except asyncio.CancelledError:
                pass

    async def on_reset(self, cancellation_token: CancellationToken) -> None:
        """Clear the chat history."""
        self._chat_history.clear()

    def _to_config(self) -> CoderAgentConfig:
        """Convert the agent's state to a configuration object."""
        return CoderAgentConfig(
            name=self.name,
            model_client=self._model_client.dump_component(),
            description=self.description,
            max_debug_rounds=self._max_debug_rounds,
            summarize_output=self._summarize_output,
            # TODO: Optionally add code_executor configuration if supported
        )

    @classmethod
    def _from_config(cls, config: CoderAgentConfig) -> Self:
        """Create an agent instance from a configuration object."""
        return cls(
            name=config.name,
            model_client=ChatCompletionClient.load_component(config.model_client),
            description=config.description,
            max_debug_rounds=config.max_debug_rounds,
            summarize_output=config.summarize_output,
            # TODO: Optionally load code_executor from config if provided
        )

    async def save_state(self) -> Mapping[str, Any]:
        """
        Save the state of the agent.
        """
        return {
            "chat_history": [msg.dump() for msg in self._chat_history],
        }

    async def load_state(self, state: Mapping[str, Any]) -> None:
        """
        Load the state of the agent.
        """
        # Create message factory for deserialization.
        message_factory = MessageFactory()
        for msg_data in state["chat_history"]:
            msg = message_factory.create(msg_data)
            assert isinstance(msg, BaseChatMessage)
            self._chat_history.append(msg)



import os
from pathlib import Path
import sys
import json
import asyncio
import argparse
import signal
from datetime import datetime
import types
import yaml
from typing import Any, Optional, Literal, cast
from autogen_core import EVENT_LOGGER_NAME, CancellationToken
from autogen_agentchat.ui import Console
from .task_team import get_task_team
from loguru import logger
import logging
from ...ui_backend.utils import LLMCallFilter
from ...ui_backend.types import RunPaths
from .magentic_ui_config import MagenticUIConfig, ModelClientConfigs

logging.basicConfig(level=logging.WARNING, handlers=[])
logger_llm = logging.getLogger(EVENT_LOGGER_NAME)
logger_llm.setLevel(logging.INFO)

ApprovalPolicy = Literal["always", "never", "auto-conservative", "auto-permissive"]


async def cancellable_input(
    prompt: str,
    cancellation_token: Optional[CancellationToken],
    input_type: Optional[str] = "text_input",
) -> str:
    # Copied from autogen_agentchat.agents.UserProxyAgent
    assert input_type in ["text_input", "approval"]
    task: asyncio.Task[str] = asyncio.create_task(asyncio.to_thread(input, prompt))
    if cancellation_token is not None:
        cancellation_token.link_future(task)
    return await task


def setup_llm_logging(log_dir: str) -> None:
    """Set up logging to capture only LLM calls to a directory."""
    log_file = os.path.join(
        log_dir, f"llm_calls_{datetime.now().strftime('%Y%m%d_%H%M%S')}.log"
    )
    file_handler = logging.FileHandler(log_file)
    file_handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(message)s"
    )  # Only log the message since it contains the JSON
    file_handler.setFormatter(formatter)
    file_handler.addFilter(LLMCallFilter())
    for handler in logger_llm.handlers[:]:  # Remove any existing handlers
        logger_llm.removeHandler(handler)
    logger_llm.addHandler(file_handler)


async def get_team(
    cooperative_planning: bool,
    autonomous_execution: bool,
    reset: bool = False,
    task: str | None = None,
    final_answer_prompt: str | None = None,
    debug: bool = False,
    state_file: str = ".magentic_ui_state.json",
    internal_workspace_root: str | None = None,
    external_workspace_root: str | None = None,
    playwright_port: int = -1,
    novnc_port: int = -1,
    inside_docker: bool = False,
    work_dir: str | None = None,
    model_context_token_limit: int = 128000,
    client_config: str | None = None,
    action_policy: ApprovalPolicy = "never",
    user_proxy_type: str | None = None,
    task_metadata: str | None = None,
    hints: str | None = None,
    answer: str | None = None,
) -> None:
    if reset:
        print(f"Resetting state file: {state_file}")
        # delete the state file if it exists
        if os.path.exists(state_file):
            print(f"Deleting state file: {state_file}")
            os.remove(state_file)

    if inside_docker:
        # Use environment variables as fallback if paths not provided
        if internal_workspace_root is None:
            internal_workspace_root = os.environ.get("INTERNAL_WORKSPACE_ROOT")
        if external_workspace_root is None:
            external_workspace_root = os.environ.get("EXTERNAL_WORKSPACE_ROOT")

        if not internal_workspace_root or not external_workspace_root:
            raise ValueError(
                "When running inside docker, both internal and external workspace root paths must be provided either via arguments or environment variables"
            )

        # In contrast to the UI -- where each session creates a new folder --
        # we need a consistent folder name so that benchmark scripts will know
        # where to place input files.
        os.makedirs(os.path.join(internal_workspace_root, "cli_files"), exist_ok=True)

        paths = RunPaths(
            internal_root_dir=Path(internal_workspace_root),
            external_root_dir=Path(external_workspace_root),
            run_suffix="run",
            internal_run_dir=Path(os.path.join(internal_workspace_root, "cli_files")),
            external_run_dir=Path(os.path.join(external_workspace_root, "cli_files")),
        )
    else:
        if not work_dir:
            raise ValueError(
                "When running outside docker, work_dir path must be provided"
            )

        work_dir_path = Path(work_dir)

        os.makedirs(os.path.join(work_dir, "cli_files"), exist_ok=True)
        work_dir_files = work_dir_path / "cli_files"

        paths = RunPaths(
            internal_root_dir=work_dir_path,
            external_root_dir=work_dir_path,
            run_suffix="run",
            internal_run_dir=work_dir_files,
            external_run_dir=work_dir_files,
        )

    client_config_dict: dict[str, Any] = {}
    if client_config:
        with open(client_config, "r") as f:
            client_config_dict = yaml.safe_load(f)

    model_client_configs = ModelClientConfigs(
        orchestrator=client_config_dict.get("orchestrator_client", None),
        web_surfer=client_config_dict.get("web_surfer_client", None),
        coder=client_config_dict.get("coder_client", None),
        file_surfer=client_config_dict.get("file_surfer_client", None),
    )

    magentic_ui_config = MagenticUIConfig(
        model_client_configs=model_client_configs,
        approval_policy=action_policy,
        cooperative_planning=cooperative_planning,
        autonomous_execution=autonomous_execution,
        allow_for_replans=True,
        do_bing_search=False,
        model_context_token_limit=model_context_token_limit,
        allow_follow_up_input=False,
        final_answer_prompt=final_answer_prompt,
        playwright_port=playwright_port,
        novnc_port=novnc_port,
        user_proxy_type=user_proxy_type,
        task=task_metadata,
        hints=hints,
        answer=answer,
        inside_docker=inside_docker,
    )

    team = await get_task_team(
        magentic_ui_config=magentic_ui_config,
        input_func=cancellable_input,
        paths=paths,
    )

    try:
        if state_file and os.path.exists(state_file) and not reset:
            state = None
            with open(state_file, "r") as f:
                state = json.load(f)
            await team.load_state(state)

        if not task:

            def flushed_input(prompt: str) -> str:
                # Prompt for input, but flush the prompt to ensure it appears immediately
                print(prompt, end="", flush=True)
                return input()

            task = await asyncio.get_event_loop().run_in_executor(
                None, flushed_input, ">: "
            )

        stream = team.run_stream(task=task)
        await Console(stream)

        state = await team.save_state()

        with open(state_file, "w") as f:
            json.dump(state, f, indent=2)

    finally:
        logger.info("Closing team...")
        await team.close()


def main() -> None:
    parser = argparse.ArgumentParser(description="Dr.Sai-UI CLI")
    parser.add_argument(
        "--disable-planning",
        dest="cooperative_planning",
        action="store_false",
        default=True,
        help="Disable co-planning mode (default: enabled), user will not be involved in the planning process",
    )
    parser.add_argument(
        "--autonomous-execution",
        dest="autonomous_execution",
        action="store_true",
        default=False,
        help="Enable autonomous execution mode (default: disabled), user will not be involved in the execution",
    )
    parser.add_argument(
        "--autonomous",
        dest="autonomous",
        action="store_true",
        default=False,
        help="Enable autonomous mode (default: disabled), no co-planning and no human involvment in execution",
    )
    parser.add_argument(
        "--reset",
        dest="reset",
        action="store_true",
        default=False,
        help="Reset the team state before running the task otherwise continue with the previous state (default: False)",
    )
    parser.add_argument(
        "--task",
        dest="task",
        type=str,
        default="",
        help="Specifies the initial task. If a plain string, use this input verbatim. If the string matches a filename, read the initial task from a file. Use '-' to read from stdin. (default: prompt's the user for the task)",
    )
    parser.add_argument(
        "--final-answer-prompt",
        dest="final_answer_prompt",
        type=str,
        default="",
        help="Overrides the final answer prompt used to summarize the conversation. If a plain string, use this input verbatim. If the string matches a filename, read the prompt from a file. (default: use orchestrator's built-in prompt)",
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Enable debug mode to show internal messages (default: disabled)",
    )
    parser.add_argument(
        "--internal-root",
        dest="internal_workspace_root",
        type=str,
        default=None,
        help="Internal workspace root directory path (default: use INTERNAL_WORKSPACE_ROOT environment variable)",
    )
    parser.add_argument(
        "--external-root",
        dest="external_workspace_root",
        type=str,
        default=None,
        help="External workspace root directory path (default: use EXTERNAL_WORKSPACE_ROOT environment variable)",
    )
    parser.add_argument(
        "--playwright-port",
        dest="playwright_port",
        type=int,
        default=-1,
        help="Port to run the Playwright browser on (default: -1 means use default port)",
    )
    parser.add_argument(
        "--novnc-port",
        dest="novnc_port",
        type=int,
        default=-1,
        help="Port to run the noVNC server on (default: -1 means use default port)",
    )
    parser.add_argument(
        "--inside-docker",
        dest="inside_docker",
        action="store_true",
        default=False,
        help="Indicates if running inside docker container (default: False)",
    )
    parser.add_argument(
        "--work-dir",
        dest="work_dir",
        type=str,
        default=None,
        help="Working directory path when running outside docker (required if not inside docker)",
    )
    parser.add_argument(
        "--config",
        dest="config",
        type=str,
        default=None,
        help="Path to the configuration file (default: 'config.yaml')",
    )
    parser.add_argument(
        "--user-proxy-type",
        dest="user_proxy_type",
        type=str,
        choices=["dummy", "metadata"],
        default=None,
        help="Type of user proxy agent to use ('dummy', 'metadata', or None for default; default: None)",
    )
    parser.add_argument(
        "--metadata-task",
        dest="metadata_task",
        type=str,
        default=None,
        help="Task description for metadata user proxy (required if user-proxy-type is 'metadata')",
    )
    parser.add_argument(
        "--metadata-hints",
        dest="metadata_hints",
        type=str,
        default=None,
        help="Task hints for metadata user proxy (required if user-proxy-type is 'metadata')",
    )
    parser.add_argument(
        "--metadata-answer",
        dest="metadata_answer",
        type=str,
        default=None,
        help="Task answer for metadata user proxy (required if user-proxy-type is 'metadata')",
    )
    parser.add_argument(
        "--llmlog-dir",
        dest="llmlog_dir",
        type=str,
        help="Directory path to save LLM call logs (if not provided, LLM logging is disabled)",
    )
    parser.add_argument(
        "--action-policy",
        dest="action_policy",
        type=str,
        default="never",
        help="ActionGuard policy ('always', 'never', 'auto-conservative', 'auto-permissive'; default: never)",
    )

    args = parser.parse_args()

    # Validate user proxy type
    if args.user_proxy_type not in [None, "dummy", "metadata"]:
        raise ValueError(
            f"Invalid user proxy type: {args.user_proxy_type}. Valid options are None, 'dummy', or 'metadata'."
        )

    # Validate metadata user proxy parameters
    if args.user_proxy_type == "metadata":
        if not all([args.metadata_task, args.metadata_hints, args.metadata_answer]):
            raise ValueError(
                "When using metadata user proxy type, all metadata parameters (--metadata-task, --metadata-hints, --metadata-answer) must be provided."
            )

    # Validate action policy
    if args.action_policy not in [
        "always",
        "never",
        "auto-conservative",
        "auto-permissive",
    ]:
        raise ValueError(
            f"Invalid action policy: {args.action_policy}. Valid options are 'always', 'never', 'auto-conservative', 'auto-permissive'."
        )

    # Set up LLM logging if requested
    if args.llmlog_dir:
        setup_llm_logging(args.llmlog_dir)

    # If the config file is not provided, check for the default config file
    client_config: str | None = args.config
    if not client_config:
        if os.path.isfile("config.yaml"):
            client_config = "config.yaml"
        else:
            logger.info("Config file not provided. Using default settings.")

    # Expand the task and final answer prompt
    task: str | None = None
    if args.task:
        if args.task == "-":
            task = sys.stdin.buffer.read().decode("utf-8")
        elif os.path.isfile(args.task):
            with open(args.task, "r") as f:
                task = f.read()
        else:
            task = args.task

    final_answer_prompt: str | None = None
    if args.final_answer_prompt:
        if os.path.isfile(args.final_answer_prompt):
            with open(args.final_answer_prompt, "r") as f:
                final_answer_prompt = f.read()
        else:
            final_answer_prompt = args.final_answer_prompt

    # Set up autonomous execution mode if requestes
    if args.autonomous:
        args.autonomous_execution = True
        args.cooperative_planning = False

    # Add a basic signal handler to log as soon as SIGINT is received
    def signal_handler(sig: int, frame: types.FrameType | None) -> Any:
        logger.info("magentic-ui cli caught SIGINT...")
        raise KeyboardInterrupt

    signal.signal(signal.SIGINT, signal_handler)

    # Run the team
    asyncio.run(
        get_team(
            cooperative_planning=args.cooperative_planning,
            autonomous_execution=args.autonomous_execution,
            reset=args.reset,
            task=task,
            final_answer_prompt=final_answer_prompt,
            debug=args.debug,
            internal_workspace_root=args.internal_workspace_root,
            external_workspace_root=args.external_workspace_root,
            playwright_port=args.playwright_port,
            novnc_port=args.novnc_port,
            inside_docker=args.inside_docker,
            work_dir=args.work_dir,
            client_config=client_config,
            action_policy=cast(ApprovalPolicy, args.action_policy),
            user_proxy_type=args.user_proxy_type,
            task_metadata=args.metadata_task
            if args.user_proxy_type == "metadata"
            else task,
            hints=args.metadata_hints if args.user_proxy_type == "metadata" else None,
            answer=args.metadata_answer if args.user_proxy_type == "metadata" else None,
        )
    )


if __name__ == "__main__":
    main()



# This Script is slightly modified from the creators of the AssistantBench dataset https://huggingface.co/spaces/AssistantBench/leaderboard/blob/main/evaluation/evaluator.py
import json
from .evaluate_factory import get_evaluator
import numpy as np


def find_isnan(samp):
    try:
        if np.isnan(samp):
            return True
        else:
            return False
    except (TypeError, ValueError):
        return False


def fix_ans(answer):
    try:
        answer = (
            answer.replace("{'", '{"')
            .replace("', '", '", "')
            .replace("': '", '": "')
            .replace("'}", '"}')
        )
        answer = answer.replace("': ", '": ')
        return answer
    except (AttributeError, TypeError):
        return answer


def parse_answer(answer):
    if len(answer) == 1:
        ans, is_num = fix_number(answer[0])
        if is_num:
            return ans, "number"
        try:
            ans = json.loads(fix_ans(answer[0]))
            return [ans], "json"
        except json.JSONDecodeError:
            ans, is_num = fix_number(answer[0])
            if is_num:
                return ans, "number"
            else:
                return answer[0], "string"
    else:
        try:
            ans = [json.loads(fix_ans(ex)) for ex in answer]
            return ans, "json"
        except json.JSONDecodeError:
            return answer, "string list"


def fix_number(number):
    if isinstance(number, str):
        copy_ans = number
        copy_ans = " ".join(
            " ".join(" ".join(copy_ans.split("$")).split("%")).split("sqft")
        ).strip()
        copy_ans = copy_ans.strip()
        copy_ans = copy_ans.replace(",", ".").replace(" square kilometers", "")
        try:
            return float(copy_ans), True
        except ValueError:
            return number, False
    elif isinstance(number, int):
        return float(number), True
    else:
        return number, True


def fix_prediction(prediction, gold_answer, evaluator):
    if (
        isinstance(prediction, list)
        and len(prediction) == 1
        and (
            isinstance(prediction[0], int)
            or (isinstance(prediction[0], str) and prediction[0].isnumeric())
        )
    ):
        prediction = fix_number(prediction[0])

    if not isinstance(prediction, list):
        prediction, is_num = fix_number(prediction)
        if evaluator == "json":
            try:
                prediction = [json.loads(pred) for pred in prediction.split("\n")]  # type: ignore
            except json.JSONDecodeError:
                prediction = [prediction]

    if (hasattr(type(prediction), "__len__")) and (len(prediction) == 0):  # type: ignore
        return prediction, False

    if (isinstance(prediction, list) and len(prediction) > 1) and isinstance(
        gold_answer, float
    ):
        return prediction, False

    return prediction, True


def ab_question_scorer(prediction, gold_answer):
    """
    prediction: str or list of str
    gold_answer: str or list of str

    returns a float between 0 and 1
    """
    try:
        try:
            prediction = json.loads(prediction)
        except json.JSONDecodeError:
            prediction = prediction

        answer_list = (
            [x for x in gold_answer.split("\n") if len(x.strip()) > 0]
            if not isinstance(gold_answer, list)
            else gold_answer
        )
        gold_answer, evaluator = parse_answer(answer_list)
        prediction, run_eval = fix_prediction(prediction, gold_answer, evaluator)

        # if (not isinstance(prediction, float) and len(prediction) == 0) or find_isnan(
        #     prediction
        # ):
        #     has_ans = 0.0

        if not run_eval:
            return 0.0

        metric_eval = get_evaluator(evaluator)
        accuracy = metric_eval(prediction, gold_answer)
        # double check if the accuracy is a number between 0 and 1
        if 0 <= accuracy <= 1:
            return accuracy
        else:
            # throw exception
            raise ValueError(
                f"Accuracy should be a float between 0 and 1, but got {accuracy}"
            )
    except Exception as e:
        print(
            f"Something went wrong while evaluating prediction {prediction} vs gold answer {gold_answer} with error {e}"
        )
        return 0.0



from ...ui_backend.input_func import InputFuncType, AsyncInputFunc, SyncInputFunc
from autogen_agentchat.messages import TextMessage, MultiModalMessage

from autogen_core import CancellationToken, Image, EVENT_LOGGER_NAME
from autogen_core.models import (
    ChatCompletionClient,
    LLMMessage,
    SystemMessage,
)

from inspect import iscoroutinefunction

import json

from contextlib import contextmanager
from contextvars import ContextVar

import asyncio
from typing import (
    Any,
    Optional,
    TypedDict,
    cast,
    ClassVar,
    Generator,
    List,
    Literal,
    Protocol,
)

from dataclasses import dataclass
import logging


"""
The ActionGuard protocol is used to check if an action is irreversisble.

The three possible values are:
- "always": The action is always irreversible.
- "maybe": The action is maybe irreversible.
- "never": The action is never irreversible.
"""
MaybeRequiresApproval = Literal["always", "maybe", "never"]

DEFAULT_REQUIRES_APPROVAL: MaybeRequiresApproval = "always"


class BaseApprovalGuard(Protocol):
    async def requires_approval(
        self,
        baseline: MaybeRequiresApproval,
        llm_guess: MaybeRequiresApproval,
        action_context: List[LLMMessage],
    ) -> bool:
        """Check if the action is irreversible; only called if the tool is marked 'maybe' irreversible."""
        ...

    async def get_approval(
        self, action_description: TextMessage | MultiModalMessage
    ) -> bool:
        """Get approval for the action; only called if the tool is marked 'always' irreversible, or
        'maybe' irreversible if the ."""
        ...


# TypedDictionary for JSON
# Scehma: { "accepted": bool, "content": str }
class UserInputResponse(TypedDict):
    accepted: bool
    content: str


@dataclass
class ApprovalConfig:
    approval_policy: Literal[
        "always", "never", "auto-conservative", "auto-permissive"
    ] = "never"


# This works around our inability to pass a callback-wrapper object through the ComponentConfig
# system, since it is designed for serialization/deserialization of the config object.
# TODO: Figure out how to include callbacks in ComponentConfig-based object initialization
class ApprovalGuardContext:
    def __init__(self) -> None:
        raise RuntimeError(
            "InputRequestContext cannot be instantiated. It is a static class that provides context management for user input requests."
        )

    _APPROVAL_GUARD_CONTEXT_VAR: ClassVar[ContextVar[BaseApprovalGuard | None]] = (
        ContextVar("_APPROVAL_GUARD_CONTEXT_VAR")
    )

    @classmethod
    @contextmanager
    def populate_context(
        cls, ctx: BaseApprovalGuard | None
    ) -> Generator[None, Any, None]:
        """:meta private:"""
        token = ApprovalGuardContext._APPROVAL_GUARD_CONTEXT_VAR.set(ctx)
        try:
            yield
        finally:
            ApprovalGuardContext._APPROVAL_GUARD_CONTEXT_VAR.reset(token)

    @classmethod
    def approval_guard(cls) -> BaseApprovalGuard | None:
        try:
            return cls._APPROVAL_GUARD_CONTEXT_VAR.get()
        except LookupError as e:
            raise RuntimeError(
                "ApprovalGuardContext.approval_guard() should only be called when initializing a WebSurfer agent during Team creation."
            ) from e


IRREVERSIBLE_CHECK_PROMPT_TEMPLATE = """
The following action is being proposed:

{approval_message}

Please review this action in the context of the following history, and determine if it requires human approval.
An action requires human approval if it is irreversible or is potentially harmful or impactful.

Please evaluate this action carefully considering the following criteria:
- Does the action have potential real-world consequences affecting user safety or security?
- Could the action cause irreversible changes or require complex steps to undo?
- Does the action impact data integrity, privacy, or significant system state?
- Could the action affect other users or external entities in any impactful way?

If the answer to any of the above is yes, the action should be considered harmful or impactful and require explicit user approval.

Examples for when the action should require human approval:
- If the websurfer is about to click a button that submits a form, or sends an email, or make a purchase, it should require human approval.
- If the websurfer is typing potentially sensitive information, it should require human approval.
- If the coder is writing code that creates files that were not approved or leaks information about the underlying system, it should require human approval.
- If the filesurfer wants to open a file that seems malicious, it should require human approval.

Examples for when the action should not require human approval:
- If the websurfer is about to click a button that changes focus in a UI, it should not require human approval.
- If the websurfer is entering a search query into a search engine that is not sensitive, it should not require human approval.
- If the coder is writing code that is not harmful or creating or opening files, it should not require human approval.

This is not an exhaustive list.

When unsure, err on the side of caution and require human approval.

Please respond with "YES" (requires human approval) or "NO" (does not require human approval) ONLY to indicate your decision.
"""


class ApprovalGuard(BaseApprovalGuard):
    def __init__(
        self,
        input_func: Optional[InputFuncType] = None,
        default_approval: bool = True,
        model_client: Optional[ChatCompletionClient] = None,
        config: Optional[ApprovalConfig] = None,
    ):
        self.model_client = model_client
        self.input_func = input_func
        self._is_async = iscoroutinefunction(input_func)
        self.default_approval = default_approval

        self.config = config or ApprovalConfig()

        self.logger = logging.getLogger(f"{EVENT_LOGGER_NAME}.ApprovalGuard")

    async def _get_input(
        self, prompt: str, cancellation_token: Optional[CancellationToken]
    ) -> str:
        """Handle input based on function signature."""
        try:
            if self._is_async:
                # Cast to AsyncInputFunc for proper typing
                async_func = cast(AsyncInputFunc, self.input_func)
                return await async_func(prompt, cancellation_token, "approval")
            else:
                # Cast to SyncInputFunc for proper typing
                sync_func = cast(SyncInputFunc, self.input_func)
                loop = asyncio.get_event_loop()
                return await loop.run_in_executor(None, sync_func, prompt, "approval")

        except asyncio.CancelledError:
            raise
        except Exception as e:
            raise RuntimeError(f"Failed to get user input: {str(e)}") from e

    async def requires_approval(
        self,
        baseline: MaybeRequiresApproval,
        llm_guess: MaybeRequiresApproval,
        action_context: list[LLMMessage],
    ) -> bool:
        if self.config.approval_policy == "always":
            return True

        if self.config.approval_policy == "never":
            return False

        # smart approval policy
        if baseline == "never":
            return False
        elif baseline == "always":
            return True

        if self.config.approval_policy == "auto-permissive":
            if llm_guess == "never":
                return False
            else:
                return True

        else:
            # auto-conservative policy mode
            # baseline == "maybe"
            if self.model_client is None:
                return self.default_approval
            action_proposal = action_context[-1] if action_context else None
            if action_proposal is None:
                return (
                    self.default_approval
                )  # TODO: Should we require an approval if we have no context?

            check_prompt = IRREVERSIBLE_CHECK_PROMPT_TEMPLATE.format(
                approval_message=action_proposal.content
            )
            system_message = SystemMessage(content=check_prompt)

            selected_context = action_context[:-1]
            if len(selected_context) > 5:
                selected_context = selected_context[-5:]

            request_messages = [system_message]

            result = await self.model_client.create(request_messages)

            if not (isinstance(result.content, str)):
                self.logger.warning(
                    "Model did not return a string response. Defaulting to True for irreversibility check."
                )
                return True

            self.logger.info(
                f"Checking action irreversibility for {action_proposal.content}:\n\n\t--- {result.content}"
            )

            if result.content.lower() in ["yes", "y"]:
                return True
            elif result.content.lower() in ["no", "n"]:
                return False
            else:
                self.logger.warning(
                    "Model did not return a valid response. Expected 'yes' or 'no'. Defaulting to True for irreversibility check."
                )
                return True

    async def get_approval(
        self, action_description: TextMessage | MultiModalMessage
    ) -> bool:
        if self.input_func is None:
            return self.default_approval

        # Use the input function to get user approval
        action_description_str: str
        if isinstance(action_description, TextMessage):
            action_description_str = action_description.content
        elif isinstance(action_description, MultiModalMessage):
            action_description_str = ""
            for content in action_description.content:
                if isinstance(content, str):
                    action_description_str += "\n" + content
                elif isinstance(content, Image):
                    action_description_str += "\n[Image]"

        result_or_json = await self._get_input(action_description_str, None)

        if isinstance(result_or_json, str):
            result_or_json = result_or_json.strip()

        # does result_or_json start with "{"?
        if result_or_json.startswith("{"):
            try:
                # { accepted: true/false, content: "..." }
                result = json.loads(result_or_json)
                if isinstance(result, dict) and "accepted" in result:
                    return bool(result["accepted"])  # type: ignore
                else:
                    return self.default_approval

            except json.JSONDecodeError:
                pass

        if result_or_json.lower() in [
            "accept",
            "yes",
            "y",
            "I don't know. Use your best judgment.",
        ]:
            return True
        elif result_or_json.lower() in ["deny", "no", "n"]:
            return False
        else:
            # If the input is not recognized, default to the default_approval
            return self.default_approval



from typing import Any, Dict, List, cast
from typing_extensions import TypedDict
from ..approval_guard import MaybeRequiresApproval
from autogen_core.tools import ToolSchema, ParametersSchema


# We define the ToolMetadata type here to avoid modifying
# the autogen_core module.
class ToolMetadata(TypedDict):
    irreversible: MaybeRequiresApproval


_tool_metadata: dict[str, ToolMetadata] = {}


def load_tool(tooldef: Dict[str, Any]) -> ToolSchema:
    tool_metadata: ToolMetadata = cast(ToolMetadata, tooldef.get("metadata", {}))
    _tool_metadata[tooldef["function"]["name"]] = tool_metadata

    return ToolSchema(
        name=tooldef["function"]["name"],
        description=tooldef["function"]["description"],
        parameters=ParametersSchema(
            type="object",
            properties=tooldef["function"]["parameters"]["properties"],
            required=tooldef["function"]["parameters"]["required"],
        ),
    )


def get_tool_metadata(tool_name_or_schema: str | ToolSchema) -> ToolMetadata:
    """Get the metadata for a tool by its name."""
    tool_name: str | None = (
        tool_name_or_schema
        if isinstance(tool_name_or_schema, str)
        else tool_name_or_schema.get("name")
    )

    metadata = _tool_metadata.get(tool_name)

    if metadata is None:
        raise ValueError(f"Tool {tool_name} not found in metadata.")

    return metadata


REQUIRE_APPROVAL_KEY = "require_approval"
REQUIRE_APPROVAL_PROMPT_FORMAT = "Is this action something that would require human approval before being done? Example: {guarded_examples}; but {unguarded_examples} are not {category}."
IRREVERSIBLE_ACTION_PROMPT_FORMAT = REQUIRE_APPROVAL_PROMPT_FORMAT


def make_approval_prompt(
    guarded_examples: List[str],
    unguarded_examples: List[str],
    category: str = "actions that require approval",
) -> str:
    """
    Create a prompt for the llm to guess as to whether the approval check is needed.
    Args:
        guarded_examples (List[str]): A list of examples of irreversible actions.
        unguarded_examples (List[str]): A list of examples of reversible actions.
        category (str, optional): The category of actions to check for approval. Default: `actions that require approval`.
    Returns:
        str: A string prompt for the actions that require approval check.
    """
    return REQUIRE_APPROVAL_PROMPT_FORMAT.format(
        guarded_examples=", ".join(guarded_examples),
        unguarded_examples=", ".join(unguarded_examples),
        category=category,
    )



from ...ui_backend.backend.cli import (
    check_docker_running,
    check_browser_image,
    check_python_image,
    build_browser_image,
    build_python_image)
import typer
from typing import Optional

def check_docker(rebuild_docker: Optional[bool] = False,):
    typer.echo("Checking if Docker is running...", nl=False)

    # Set things up for Docker
    typer.echo("Checking if Docker is running...", nl=False)

    if not check_docker_running():
        typer.echo(typer.style("Failed\n", fg=typer.colors.RED, bold=True))
        typer.echo("Docker is not running. Please start Docker and try again.")
        raise typer.Exit(1)
    else:
        typer.echo(typer.style("OK", fg=typer.colors.GREEN, bold=True))

    typer.echo("Checking Docker vnc browser image...", nl=False)
    if not check_browser_image() or rebuild_docker:
        typer.echo(typer.style("Update\n", fg=typer.colors.YELLOW, bold=True))
        typer.echo("Building Docker vnc image (this WILL take a few minutes)")
        build_browser_image()
        typer.echo("\n")
    else:
        typer.echo(typer.style("OK", fg=typer.colors.GREEN, bold=True))

    typer.echo("Checking Docker python image...", nl=False)
    if not check_python_image() or rebuild_docker:
        typer.echo(typer.style("Update\n", fg=typer.colors.YELLOW, bold=True))
        typer.echo("Building Docker python image (this WILL take a few minutes)")
        build_python_image()
        typer.echo("\n")
    else:
        typer.echo(typer.style("OK", fg=typer.colors.GREEN, bold=True))

    # check the images again and throw an error if they are not found
    if not check_browser_image() or not check_python_image():
        typer.echo(typer.style("Failed\n", fg=typer.colors.RED, bold=True))
        typer.echo("Docker images not found. Please build the images and try again.")
        raise typer.Exit(1)




import json
from typing import Union, List, Optional
from autogen_agentchat.messages import TextMessage, MultiModalMessage
from autogen_core.models import ChatCompletionClient
from autogen_core.models import LLMMessage, UserMessage
from autogen_core.model_context import TokenLimitedChatCompletionContext
from ..types import Plan


def chat_msg_to_llm_message(
    message: Union[TextMessage, MultiModalMessage],
) -> LLMMessage:
    if isinstance(message, TextMessage):
        if isinstance(message.content, str):
            return UserMessage(
                content=[message.content],
                source="user" if isinstance(message, TextMessage) else "magentic-ui",
            )
        else:
            raise ValueError(f"Unsupported content type: {type(message.content)}")
    elif isinstance(message, MultiModalMessage):
        return UserMessage(content=message.content, source="magentic-ui")
    else:
        raise ValueError(f"Unsupported message type: {type(message)}")


async def learn_plan_from_messages(
    client: ChatCompletionClient,
    messages: List[Union[TextMessage, MultiModalMessage]],
) -> Plan:
    """
    Given a sequence of chat messages, use structured outputs to create a draft of parameterized plan.

    Args:
        client (ChatCompletionClient): The chat completion client to use for generating the plan.
        messages (List[TextMessage | MultiModalMessage]): A list of chat messages to learn the plan from.

    Returns:
        Plan: The learned plan.
    """
    llm_messages: List[LLMMessage] = []
    for message in messages:
        llm_message = chat_msg_to_llm_message(message)
        llm_messages.append(llm_message)

    instruction_message = UserMessage(
        content=[
            """
The above messages are a conversation between a user and an AI assistant.
The AI assistant helped the user with their task and arrived potentially at a "Final Answer" to accomplish their task.

We want to be able to learn a plan from the conversation that can be used to accomplish the task as efficiently as possible.
This plan should help us accomplish this task and tasks similar to it more efficiently in the future as we learned from the mistakes and successes of the AI assistant and the details of the conversation.

Guidelines:
- We want the most efficient and direct plan to accomplish the task. The less number of steps, the better. Some agents can perform multiple steps in one go.
- We don't need to repeat the exact sequence of the conversation, but rather we need to focus on how to get to the final answer most efficiently without directly giving the final answer.
- Include details about the actions performed, buttons clicked, urls visited if they are useful.
For instance, if the plan was trying to find the github stars of autogen and arrived at the link https://github.com/microsoft/autogen then mention that link.
Or if the web surfer clicked a specific button to create an issue, mention that button.

Here is an example of a plan that the AI assistant might follow:

Example:

User request: "On which social media platform does Autogen have the most followers?"

Step 1:
- title: "Find all social media platforms that Autogen is on"
- details: "1) do a search for autogen social media platforms using Bing, 2) find the official link for autogen where the social media platforms might be listed, 3) report back all the social media platforms that Autogen is on"
- agent_name: "web_surfer"

Step 2:
- title: "Find the number of followers on Twitter"
- details: "Go to the official link for autogen on the web and find the number of followers on Twitter"
- agent_name: "web_surfer"

Step 3:
- title: "Find the number of followers on LinkedIn"
- details: "Go to the official link for autogen on the web and find the number of followers on LinkedIn"
- agent_name: "web_surfer"

Please provide the plan from the conversation above. Again, DO NOT memorize the final answer in the plan.
            """
        ],
        source="user",
    )

    # Create token limited context
    model_context = TokenLimitedChatCompletionContext(client, token_limit=110000)
    await model_context.clear()
    await model_context.add_message(instruction_message)
    for msg in llm_messages:
        await model_context.add_message(msg)
    token_limited_messages = await model_context.get_messages()

    response = await client.create(
        messages=token_limited_messages,
        extra_create_args={"response_format": Plan},
    )

    response_content: Optional[str] = (
        response.content if isinstance(response.content, str) else None
    )

    if response_content is None:
        raise ValueError("Response content is not a valid JSON string")

    plan = Plan.model_validate(json.loads(response_content))

    return plan


async def adapt_plan(client: ChatCompletionClient, plan: Plan, task: str) -> Plan:
    """
    Given a plan and new task, adapt the plan to the new task.

    Args:
        client (ChatCompletionClient): The chat completion client to use for adapting the plan.
        plan (Plan): The plan to adapt.
        task (str): The new task to adapt the plan to.

    Returns:
        Plan: The adapted plan.
    """

    instruction_message = UserMessage(
        content=["Adapt the following plan to the new task."],
        source="user",
    )
    plan_message = UserMessage(
        content=[json.dumps(plan.model_dump())],
        source="user",
    )
    task_message = UserMessage(
        content=[task],
        source="user",
    )

    response = await client.create(
        messages=[instruction_message, plan_message, task_message],
        extra_create_args={"response_format": Plan},
    )

    response_content: Optional[str] = (
        response.content if isinstance(response.content, str) else None
    )

    if response_content is None:
        raise ValueError("Response content is not a valid JSON string")

    adapted_plan = Plan.model_validate(json.loads(response_content))

    return adapted_plan



import json
import re
from typing import Any, Optional


def is_accepted_str(user_input: str) -> bool:
    LIST_OF_ACCEPTED_STRS = [
        "accept",
        "accepted",
        "acept",
        "run",
        "execute plan",
        "execute",
        "looks good",
        "do it",
        "accept plan",
        "accpt",
        "run plan",
        "sounds good",
        "i don't know. use your best judgment.",
        "i don't know, you figure it out, don't ask me again.",
    ]
    user_input = user_input.lower()
    user_input = user_input.strip()
    if user_input in LIST_OF_ACCEPTED_STRS:
        return True
    return False


def extract_json_from_string(s: str) -> Optional[Any]:
    """
    Searches for a JSON object within the string and returns the loaded JSON if found, otherwise returns None.
    """
    # Regex to find JSON objects (greedy, matches first { to last })
    match = re.search(r"\{.*\}", s, re.DOTALL)
    if match:
        json_str = match.group(0)
        try:
            return json.loads(json_str)
        except json.JSONDecodeError:
            return None
    return None



from __future__ import annotations

import asyncio
import logging

from pathlib import Path
import secrets
import socket

from autogen_core import Component
import docker
from docker.models.containers import Container
from pydantic import BaseModel

from .base_playwright_browser import DockerPlaywrightBrowser


# Configure logging
logger = logging.getLogger(__name__)


class VncDockerPlaywrightBrowserConfig(BaseModel):
    """
    Configuration for the VNC Docker Playwright Browser.
    """

    bind_dir: Path
    image: str = "magentic-ui-vnc-browser"
    playwright_port: int = 37367
    novnc_port: int = 6080
    playwright_websocket_path: str | None = None
    inside_docker: bool = True


class VncDockerPlaywrightBrowser(
    DockerPlaywrightBrowser, Component[VncDockerPlaywrightBrowserConfig]
):
    """
    A Docker-based Playwright browser implementation with VNC support for visual interaction.
    Provides both programmatic browser control via Playwright and visual access through noVNC.

    Args:
        bind_dir (Path): Directory to bind mount into the container for file access.
        image (str, optional): Docker image name for the VNC-enabled browser. Default: "magentic-ui-vnc-browser".
        playwright_port (int, optional): Port for Playwright WebSocket connection. Default: 37367.
        playwright_websocket_path (str | None, optional): Custom WebSocket path. If None, generates random path.
        novnc_port (int, optional): Port for noVNC web interface. Default: 6080.
        inside_docker (bool, optional): Whether the client is running inside Docker. Default: True.
        network_name (str, optional): Docker network name for container communication. Default: `my-network`.

    Properties:
        browser_address (str): WebSocket URL for Playwright connection.
        vnc_address (str): HTTP URL for noVNC web interface.

    Example:
        ```python
        browser = VncDockerPlaywrightBrowser(
            bind_dir=Path("./workspace"),
            playwright_port=37367,
            novnc_port=6080
        )
        await browser.start()
        # Access browser programmatically via Playwright
        # Access visual interface via noVNC at browser.vnc_address
        await browser.close()
        ```

    Note:
        Requires the Docker image 'magentic-ui-vnc-browser' to be available locally.
        Build using the Dockerfile in docker/browser-docker directory.
    """

    component_config_schema = VncDockerPlaywrightBrowserConfig
    component_type = "other"

    def __init__(
        self,
        *,
        bind_dir: Path,
        image: str = "magentic-ui-vnc-browser",
        playwright_port: int = 37367,
        playwright_websocket_path: str | None = None,
        novnc_port: int = 6080,
        inside_docker: bool = True,
        network_name: str = "my-network",
    ):
        super().__init__()
        self._bind_dir = bind_dir
        self._image = image
        self._playwright_port = playwright_port
        self._novnc_port = novnc_port
        self._playwright_websocket_path = (
            playwright_websocket_path or secrets.token_hex(16)
        )
        self._inside_docker = inside_docker
        self._network_name = network_name
        self._hostname = (
            f"magentic-ui-vnc-browser_{self._playwright_websocket_path}_{self._novnc_port}"
            if inside_docker
            else "localhost"
        )
        self._docker_name = f"magentic-ui-vnc-browser_{self._playwright_websocket_path}_{self._novnc_port}"

    def _get_available_port(self) -> tuple[int, socket.socket]:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(("127.0.0.1", 0))
        port = s.getsockname()[1]
        return port, s

    # TODO: This is a temporary solution to avoid port conflicts. Ideally we should allow docker to tell us which sockets to use
    def _generate_new_browser_address(self) -> None:
        """
        Generate new ports for Playwright and noVNC.
        """
        self._playwright_port, playwright_sock = self._get_available_port()
        self._novnc_port, novnc_sock = self._get_available_port()
        self._hostname = (
            f"magentic-ui-vnc-browser_{self._playwright_websocket_path}_{self._novnc_port}"
            if self._inside_docker
            else "localhost"
        )
        self._docker_name = f"magentic-ui-vnc-browser_{self._playwright_websocket_path}_{self._novnc_port}"
        playwright_sock.close()
        novnc_sock.close()

    @property
    def browser_address(self) -> str:
        """
        Get the address of the Playwright browser.
        """
        return f"ws://{self._hostname}:{self._playwright_port}/{self._playwright_websocket_path}"

    @property
    def vnc_address(self) -> str:
        """
        Get the address of the noVNC server.
        """
        return f"http://{self._hostname}:{self._novnc_port}/vnc.html"

    @property
    def novnc_port(self) -> int:
        """
        Get the address of the noVNC server.
        """
        return self._novnc_port

    @property
    def playwright_port(self) -> int:
        """
        Get the address of the noVNC server.
        """
        return self._playwright_port

    async def create_container(self) -> Container:
        """
        Start a headless Playwright browser using the official Playwright Docker image.
        """
        logger.info(
            f"Starting headless Playwright browser on port {self._playwright_port}..."
        )

        client = docker.from_env()

        return await asyncio.to_thread(
            client.containers.create,
            name=self._docker_name,
            image=self._image,
            detach=True,
            auto_remove=True,
            network=self._network_name if self._inside_docker else None,
            ports={
                f"{self._playwright_port}/tcp": self._playwright_port,
                f"{self._novnc_port}/tcp": self._novnc_port,
            },
            volumes={
                str(self._bind_dir.resolve()): {"bind": "/workspace", "mode": "rw"}
            },
            environment={
                "PLAYWRIGHT_WS_PATH": self._playwright_websocket_path,
                "PLAYWRIGHT_PORT": str(self._playwright_port),
                "NO_VNC_PORT": str(self._novnc_port),
            },
            security_opt=["seccomp=unconfined"],  # <--- 添加这一行
        )

    def _to_config(self) -> VncDockerPlaywrightBrowserConfig:
        return VncDockerPlaywrightBrowserConfig(
            bind_dir=self._bind_dir,
            image=self._image,
            playwright_port=self._playwright_port,
            novnc_port=self._novnc_port,
            playwright_websocket_path=self._playwright_websocket_path,
            inside_docker=self._inside_docker,
        )

    @classmethod
    def _from_config(
        cls, config: VncDockerPlaywrightBrowserConfig
    ) -> VncDockerPlaywrightBrowser:
        return cls(
            bind_dir=config.bind_dir,
            image=config.image,
            playwright_port=config.playwright_port,
            novnc_port=config.novnc_port,
            playwright_websocket_path=config.playwright_websocket_path,
            inside_docker=config.inside_docker,
        )



from urllib.parse import quote_plus
from playwright.async_api import async_playwright
from urllib.parse import urlparse
import asyncio
import tiktoken
from dataclasses import dataclass
from loguru import logger
from ..tools import PlaywrightController


@dataclass
class BingSearchResults:
    search_results: str
    links: list[dict[str, str]]
    page_contents: dict[str, str]
    combined_content: str


async def extract_page_markdown(url: str) -> tuple[str, str]:
    """Extract markdown content from a given URL.

    Args:
        url (str): The URL to extract content from

    Returns:
        A tuple containing:
            - str: The URL
            - str: The markdown content extracted from the page
    """
    try:
        async with async_playwright() as p:
            launch_args = ["--disable-extensions", "--disable-file-system"]
            browser = await p.chromium.launch(
                headless=True, env={}, args=launch_args, chromium_sandbox=True
            )
            context = await browser.new_context(
                accept_downloads=False,  # Disable downloads
                permissions=[],  # No additional permissions
            )

            controller = PlaywrightController()

            page = await context.new_page()
            try:
                await page.goto(url)
                await asyncio.sleep(1)
                markdown = await controller.get_page_markdown(page)
            except Exception as e:
                logger.error(f"Error extracting content: {e}")
                markdown = "Error extracting content"

            await context.close()
            await browser.close()
            return url, markdown
    except Exception as e:
        logger.error(f"Error extracting content: {e}")
        return url, "Error extracting content"


async def get_bing_search_results(
    query: str,
    max_pages: int = 3,
    timeout_seconds: int = 10,
    max_tokens_per_page: int = 10000,
) -> BingSearchResults:
    """Get the Bing search results for a given query.

    WARNING: This function spawns multiple browser local playwright instances which may consume a lot of resources and can cause risks.

    Args:
        query (str): The search query to use
        max_pages (int, optional): Maximum number of pages to extract. Default: 3
        timeout_seconds (int, optional): Maximum time in seconds to wait for search results. Default: 10
        max_tokens_per_page (int, optional): Maximum number of tokens to extract from each page. Default: 10000

    Returns:
        BingSearchResults: Contains search results markdown, links, and extracted content
    """
    search_results: str = ""
    links: list[dict[str, str]] = []
    page_contents: dict[str, str] = {}
    combined_content: str = ""

    try:
        result = await asyncio.wait_for(
            extract_page_markdown(
                f"https://www.bing.com/search?q={quote_plus(query)}&FORM=QBLH"
            ),
            timeout=timeout_seconds,
        )
        _, search_results = result

        # Extract links from markdown
        def extract_links(markdown_text: str) -> list[dict[str, str]]:
            """Extract links from markdown text.

            Args:
                markdown_text (str): The markdown text to extract links from

            Returns:
                list[dict[str, str]]: List of dictionaries containing display_text and url
            """

            def is_valid_url(url: str) -> bool:
                """Check if a URL is valid."""
                try:
                    result = urlparse(url)
                    return all([result.scheme in ("http", "https"), result.netloc])
                except Exception:
                    return False

            links: list[dict[str, str]] = []
            lines = markdown_text.split("\n")
            for line in lines:
                # Match markdown link format: [display_text](url)
                if (
                    line.count("[") == 1
                    and line.count("]") == 1
                    and line.count("(") == 1
                    and line.count(")") == 1
                ):
                    display_start = line.find("[") + 1
                    display_end = line.find("]")
                    url_start = line.find("(") + 1
                    url_end = line.find(")")

                    if all(
                        i != -1
                        for i in [
                            display_start,
                            display_end,
                            url_start,
                            url_end,
                        ]
                    ):
                        display_text = line[display_start:display_end]
                        url = line[url_start:url_end]

                        # Only add if URL is valid
                        if is_valid_url(url):
                            links.append({"display_text": display_text, "url": url})
            return links

        links = extract_links(search_results)

        # Extract content from first 5 links in parallel
        first_few_urls = [link["url"] for link in links[:max_pages]]
        tasks = [extract_page_markdown(url) for url in first_few_urls]
        extracted_contents = await asyncio.gather(*tasks, return_exceptions=True)

        # Create a dictionary mapping URLs to their content, handling any failed extractions
        page_contents: dict[str, str] = {}
        for i, result in enumerate(extracted_contents):
            url = first_few_urls[i]
            if isinstance(result, Exception):
                continue
            elif isinstance(result, tuple):
                _, content = result  # type: ignore
                if content != "Error extracting content":
                    page_contents[url] = content

        # Combine all extracted page contents into a single string
        if page_contents:
            combined_content = "Search Results for " + query + "\n\n"
            for url, content in page_contents.items():
                if max_tokens_per_page == -1:
                    token_limited_content = content
                else:
                    tokenizer = tiktoken.encoding_for_model("gpt-4o")
                    tokens = tokenizer.encode(content)
                    limited_content = tokenizer.decode(tokens[:max_tokens_per_page])
                    token_limited_content = limited_content
                combined_content += f"Page: {url}\n{token_limited_content}\n\n"

    except asyncio.TimeoutError as e:
        logger.info(f"Timeout error: {e}")
        # If we timeout, we'll still return any partial results we managed to get
        if not combined_content and page_contents:
            # If we got some pages but hadn't combined them yet
            combined_content = (
                "Search Results for " + query + " (Partial results due to timeout)\n\n"
            )
            for url, content in page_contents.items():
                if max_tokens_per_page == -1:
                    token_limited_content = content
                else:
                    tokenizer = tiktoken.encoding_for_model("gpt-4o")
                    tokens = tokenizer.encode(content)
                    limited_content = tokenizer.decode(tokens[:max_tokens_per_page])
                    token_limited_content = limited_content
                combined_content += f"Page: {url}\n{token_limited_content}\n\n"
        elif not search_results:
            # If we got absolutely nothing, return empty results
            return BingSearchResults("", [], {}, "")
    except Exception as e:
        logger.error(f"Error getting Bing search results: {e}")
        return BingSearchResults("", [], {}, "")
    return BingSearchResults(search_results, links, page_contents, combined_content)



from typing import Any, Dict, List

# ORCHESTRATOR_SYSTEM_MESSAGE_PLANNING = """
# You are a helpful AI assistant named Magentic-UI built by Microsoft Research AI Frontiers.
ORCHESTRATOR_SYSTEM_MESSAGE_PLANNING = """
You are a helpful AI assistant named Dr.Sai built by HEPAI.
Your goal is to help the user with their request.
You can complete actions on the web, complete actions on behalf of the user, execute code, and more.
You have access to a team of agents who can help you answer questions and complete tasks.
The browser the web_surfer accesses is also controlled by the user.
You are primarly a planner, and so you can devise a plan to do anything. 


The date today is: {date_today}


First consider the following:

- is the user request missing information and can benefit from clarification? For instance, if the user asks "book a flight", the request is missing information about the destination, date and we should ask for clarification before proceeding. Do not ask to clarify more than once, after the first clarification, give a plan.
- is the user request something that can be answered from the context of the conversation history without executing code, or browsing the internet or executing other tools? If so, we should answer the question directly in as much detail as possible.


Case 1: If the above is true, then we should provide our answer in the "response" field and set "needs_plan" to False.

Case 2: If the above is not true, then we should consider devising a plan for addressing the request. If you are unable to answer a request, always try to come up with a plan so that other agents can help you complete the task.


For Case 2:

You have access to the following team members that can help you address the request each with unique expertise:

{team}


Your plan should should be a sequence of steps that will complete the task.

Each step should have a title and details field.

The title should be a short one sentence description of the step.

The details should be a detailed description of the step. The details should be written in first person and should be phrased as if you are directly talking to the user.
The first sentence of the details should recap the title. We then follow it with a new line. We then add the details of the step without repeating information of the title. We should be concise but mention all crucial details to allow the human to verify the step.

Example 1:

User request: "Report back the menus of three restaurants near the zipcode 98052"

Step 1:
- title: "Locate the menu of the first restaurant"
- details: "I will locate the menu of the first restaurant. \n To accomplish this, I will search for highly-rated restaurants in the 98052 area using Bing, select one with good reviews and an accessible menu, then extract and format the menu information for reporting."
- agent_name: "web_surfer"

Step 2:
- title: "Locate the menu of the second restaurant"
- details: "I will locate the menu of the second restaurant. \n After excluding the first restaurant, I will search for another well-reviewed establishment in 98052, ensuring it has a different cuisine type for variety, then collect and format its menu information."
- agent_name: "web_surfer"

Step 3:
- title: "Locate the menu of the third restaurant"
- details: "I will locate the menu of the third restaurant. \n Building on the previous searches but excluding the first two restaurants, I will find a third establishment with a distinct cuisine type, verify its menu is available online, and compile the menu details."
- agent_name: "web_surfer"




Example 2:

User request: "Execute the starter code for the autogen repo"

Step 1:
- title: "Locate the starter code for the autogen repo"
- details: "I will locate the starter code for the autogen repo. \n This involves searching for the official AutoGen repository on GitHub, navigating to their examples or getting started section, and identifying the recommended starter code for new users."
- agent_name: "web_surfer"

Step 2:
- title: "Execute the starter code for the autogen repo"
- details: "I will execute the starter code for the autogen repo. \n This requires setting up the Python environment with the correct dependencies, ensuring all required packages are installed at their specified versions, and running the starter code while capturing any output or errors."
- agent_name: "coder_agent"


Example 3:

User request: "On which social media platform does Autogen have the most followers?"

Step 1:
- title: "Find all social media platforms that Autogen is on"
- details: "I will find all social media platforms that Autogen is on. \n This involves searching for AutoGen's official presence across major platforms like GitHub, Twitter, LinkedIn, and others, then compiling a comprehensive list of their verified accounts."
- agent_name: "web_surfer"

Step 2:
- title: "Find the number of followers for each social media platform"
- details: "I will find the number of followers for each social media platform. \n For each platform identified, I will visit AutoGen's official profile and record their current follower count, ensuring to note the date of collection for accuracy."
- agent_name: "web_surfer"

Step 3:
- title: "Find the number of followers for the remaining social media platform that Autogen is on"
- details: "For each of the remaining social media platforms that Autogen is on, find the number of followers. \n This involves visiting the remaining platforms and recording their follower counts."
- agent_name: "web_surfer"



Example 4:

User request: "Can you paraphrase the following sentence: 'The quick brown fox jumps over the lazy dog'"

You should not provide a plan for this request. Instead, just answer the question directly.


Helpful tips:
- If the plan needs information from the user, try to get that information before creating the plan.
- When creating the plan you only need to add a step to the plan if it requires a different agent to be completed, or if the step is very complicated and can be split into two steps.
- Remember, there is no requirement to involve all team members -- a team member's particular expertise may not be needed for this task.
- Aim for a plan with the least number of steps possible.
- Use a search engine or platform to find the information you need. For instance, if you want to look up flight prices, use a flight search engine like Bing Flights. However, your final answer should not stop with a Bing search only.
- If there are images attached to the request, use them to help you complete the task and describe them to the other agents in the plan.


"""


# ORCHESTRATOR_SYSTEM_MESSAGE_PLANNING_AUTONOMOUS = """
# You are a helpful AI assistant named Magentic-UI built by Microsoft Research AI Frontiers.
ORCHESTRATOR_SYSTEM_MESSAGE_PLANNING_AUTONOMOUS = """
You are a helpful AI assistant named Dr.Sai built by HEPAI.
Your goal is to help the user with their request.
You can complete actions on the web, complete actions on behalf of the user, execute code, and more.
You have access to a team of agents who can help you answer questions and complete tasks.
You are primarly a planner, and so you can devise a plan to do anything. 

The date today is: {date_today}



You have access to the following team members that can help you address the request each with unique expertise:

{team}



Your plan should should be a sequence of steps that will complete the task.

Each step should have a title and details field.

The title should be a short one sentence description of the step.

The details should be a detailed description of the step. The details should be written in first person and should be phrased as if you are directly talking to the user.
The first sentence of the details should recap the title. We then follow it with a new line. We then add the details of the step without repeating information of the title. We should be concise but mention all crucial details to allow the human to verify the step.


Example 1:

User request: "Report back the menus of three restaurants near the zipcode 98052"

Step 1:
- title: "Locate the menu of the first restaurant"
- details: "I will locate the menu of the first restaurant. \n This involves searching for top-rated restaurants in the 98052 area, selecting one with good reviews and an accessible menu, then extracting and formatting the menu information."
- agent_name: "web_surfer"

Step 2:
- title: "Locate the menu of the second restaurant"
- details: "I will locate the menu of the second restaurant. \n After excluding the first restaurant, I will search for another well-reviewed establishment in 98052, ensuring it has a different cuisine type for variety, then collect and format its menu information."
- agent_name: "web_surfer"

Step 3:
- title: "Locate the menu of the third restaurant"
- details: "I will locate the menu of the third restaurant. \n Building on the previous searches but excluding the first two restaurants, I will find a third establishment with a distinct cuisine type, verify its menu is available online, and compile the menu details."
- agent_name: "web_surfer"



Example 2:

User request: "Execute the starter code for the autogen repo"

Step 1:
- title: "Locate the starter code for the autogen repo"
- details: "I will locate the starter code for the autogen repo. \n This involves searching for the official AutoGen repository on GitHub, navigating to their examples or getting started section, and identifying the recommended starter code for new users."
- agent_name: "web_surfer"

Step 2:
- title: "Execute the starter code for the autogen repo"
- details: "I will execute the starter code for the autogen repo. \n This requires setting up the Python environment with the correct dependencies, ensuring all required packages are installed at their specified versions, and running the starter code while capturing any output or errors."
- agent_name: "coder_agent"



Example 3:

User request: "On which social media platform does Autogen have the most followers?"

Step 1:
- title: "Find all social media platforms that Autogen is on"
- details: "I will find all social media platforms that Autogen is on. \n This involves searching for AutoGen's official presence across major platforms like GitHub, Twitter, LinkedIn, and others, then compiling a comprehensive list of their verified accounts."
- agent_name: "web_surfer"

Step 2:
- title: "Find the number of followers for each social media platform"
- details: "I will find the number of followers for each social media platform. \n For each platform identified, I will visit AutoGen's official profile and record their current follower count, ensuring to note the date of collection for accuracy."
- agent_name: "web_surfer"

Step 3:
- title: "Find the number of followers for the remaining social media platform that Autogen is on"
- details: "For each of the remaining social media platforms that Autogen is on, find the number of followers. \n This involves visiting the remaining platforms and recording their follower counts."
- agent_name: "web_surfer"



Helpful tips:
- When creating the plan you only need to add a step to the plan if it requires a different agent to be completed, or if the step is very complicated and can be split into two steps.
- Aim for a plan with the least number of steps possible.
- Use a search engine or platform to find the information you need. For instance, if you want to look up flight prices, use a flight search engine like Bing Flights. However, your final answer should not stop with a Bing search only.
- If there are images attached to the request, use them to help you complete the task and describe them to the other agents in the plan.

"""


ORCHESTRATOR_PLAN_PROMPT_JSON = """
You have access to the following team members that can help you address the request each with unique expertise:

{team}

Remember, there is no requirement to involve all team members -- a team member's particular expertise may not be needed for this task.


{additional_instructions}



Your plan should should be a sequence of steps that will complete the task.

Each step should have a title and details field.

The title should be a short one sentence description of the step.

The details should be a detailed description of the step. The details should be written in first person and should be phrased as if you are directly talking to the user.
The first sentence of the details should recap the title in one short sentence. We then follow it with a new line. We then add the details of the step without repeating information of the title. We should be concise but mention all crucial details to allow the human to verify the step.
The details should not be longer that 2 sentences.


Output an answer in pure JSON format according to the following schema. The JSON object must be parsable as-is. DO NOT OUTPUT ANYTHING OTHER THAN JSON, AND DO NOT DEVIATE FROM THIS SCHEMA:

The JSON object should have the following structure



{{
"response": "a complete response to the user request for Case 1.",
"task": "a complete description of the task requested by the user",
"plan_summary": "a complete summary of the plan if a plan is needed, otherwise an empty string",
"needs_plan": boolean,
"steps":
[
{{
    "title": "title of step 1",
    "details": " rephrase the title in one short sentence \n remaining details of step 1",
    "agent_name": "the name of the agent that should complete the step"
}},
{{
    "title": "title of step 2",
    "details": " rephrase the title in one short sentence \n remaining details of step 2",
    "agent_name": "the name of the agent that should complete the step"
}},
...
]
}}
"""


ORCHESTRATOR_PLAN_REPLAN_JSON = (
    """

The task we are trying to complete is:

{task}

The plan we have tried to complete is:

{plan}

We have not been able to make progress on our task.

We need to find a new plan to tackle the task that addresses the failures in trying to complete the task previously.
"""
    + ORCHESTRATOR_PLAN_PROMPT_JSON
)


# ORCHESTRATOR_SYSTEM_MESSAGE_EXECUTION = """
# You are a helpful AI assistant named Magentic-UI built by Microsoft Research AI Frontiers.
ORCHESTRATOR_SYSTEM_MESSAGE_EXECUTION = """
You are a helpful AI assistant named Dr.Sai built by HEPAI.
Your goal is to help the user with their request.
You can complete actions on the web, complete actions on behalf of the user, execute code, and more.
The browser the web_surfer accesses is also controlled by the user.
You have access to a team of agents who can help you answer questions and complete tasks.

The date today is: {date_today}
"""


ORCHESTRATOR_PROGRESS_LEDGER_PROMPT = """
Recall we are working on the following request:

{task}

This is our current plan:

{plan}

We are at step index {step_index} in the plan which is 

Title: {step_title}

Details: {step_details}

agent_name: {agent_name}

And we have assembled the following team:

{team}

The browser the web_surfer accesses is also controlled by the user.


To make progress on the request, please answer the following questions, including necessary reasoning:

    - is_current_step_complete: Is the current step complete? (True if complete, or False if the current step is not yet complete)
    - need_to_replan: Do we need to create a new plan? (True if user has sent new instructions and the current plan can't address it. True if the current plan cannot address the user request because we are stuck in a loop, facing significant barriers, or the current approach is not working. False if we can continue with the current plan. Most of the time we don't need a new plan.)
    - instruction_or_question: Provide complete instructions to accomplish the current step with all context needed about the task and the plan. Provide a very detailed reasoning chain for how to complete the step. If the next agent is the user, pose it directly as a question. Otherwise pose it as something you will do.
    - agent_name: Decide which team member should complete the current step from the list of team members: {names}. 
    - progress_summary: Summarize all the information that has been gathered so far that would help in the completion of the plan including ones not present in the collected information. This should include any facts, educated guesses, or other information that has been gathered so far. Maintain any information gathered in the previous steps.

Important: it is important to obey the user request and any messages they have sent previously.

{additional_instructions}

Please output an answer in pure JSON format according to the following schema. The JSON object must be parsable as-is. DO NOT OUTPUT ANYTHING OTHER THAN JSON, AND DO NOT DEVIATE FROM THIS SCHEMA:

    {{
        "is_current_step_complete": {{
            "reason": string,
            "answer": boolean
        }},
        "need_to_replan": {{
            "reason": string,
            "answer": boolean
        }},
        "instruction_or_question": {{
            "answer": string,
            "agent_name": string (the name of the agent that should complete the step from {names})
        }},
        "progress_summary": "a summary of the progress made so far"

    }}
"""


ORCHESTRATOR_FINAL_ANSWER_PROMPT = """
We are working on the following task:
{task}


The above messages contain the steps that took place to complete the task.

Based on the information gathered, provide a final response to the user in response to the task.

Make sure the user can easily verify your answer, include links if there are any.

There is no need to be verbose, but make sure it contains enough information for the user.
"""

INSTRUCTION_AGENT_FORMAT = """
Step {step_index}: {step_title}
\n\n
{step_details}
\n\n
Instruction for {agent_name}: {instruction}
"""


ORCHESTRATOR_TASK_LEDGER_FULL_FORMAT = """
We are working to address the following user request:
\n\n
{task}
\n\n
To answer this request we have assembled the following team:
\n\n
{team}
\n\n
Here is the plan to follow as best as possible:
\n\n
{plan}
"""


def validate_ledger_json(json_response: Dict[str, Any], agent_names: List[str]) -> bool:
    required_keys = [
        "is_current_step_complete",
        "need_to_replan",
        "instruction_or_question",
        "progress_summary",
    ]

    if not isinstance(json_response, dict):
        return False

    for key in required_keys:
        if key not in json_response:
            return False

    # Check structure of boolean response objects
    for key in [
        "is_current_step_complete",
        "need_to_replan",
    ]:
        if not isinstance(json_response[key], dict):
            return False
        if "reason" not in json_response[key] or "answer" not in json_response[key]:
            return False

    # Check instruction_or_question structure
    if not isinstance(json_response["instruction_or_question"], dict):
        return False
    if (
        "answer" not in json_response["instruction_or_question"]
        or "agent_name" not in json_response["instruction_or_question"]
    ):
        return False
    if json_response["instruction_or_question"]["agent_name"] not in agent_names:
        return False

    # Check progress_summary is a string
    if not isinstance(json_response["progress_summary"], str):
        return False

    return True


def validate_plan_json(json_response: Dict[str, Any]) -> bool:
    if not isinstance(json_response, dict):
        return False
    required_keys = ["task", "steps", "needs_plan", "response", "plan_summary"]
    for key in required_keys:
        if key not in json_response:
            return False
    plan = json_response["steps"]
    for item in plan:
        if not isinstance(item, dict):
            return False
        if "title" not in item or "details" not in item or "agent_name" not in item:
            return False
    return True



# defines how core data types are serialized and stored in the database

from datetime import datetime
from enum import Enum
from typing import Any, List, Optional, Union

from autogen_core import ComponentModel
from pydantic import field_serializer
from sqlalchemy import ForeignKey, Integer
from sqlmodel import JSON, Column, DateTime, Field, SQLModel, func

from .types import (
    GalleryConfig,
    MessageConfig,
    MessageMeta,
    SettingsConfig,
    TeamResult,
    GalleryComponents,
    GalleryMetadata,
    AgentModeSetting,

)


class Team(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    user_id: Optional[str] = None
    version: Optional[str] = "0.0.1"
    component: Union[ComponentModel, dict[str, Any]] = Field(sa_column=Column(JSON))


class Message(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    user_id: Optional[str] = None
    version: Optional[str] = "0.0.1"
    config: Union[MessageConfig, dict[str, Any]] = Field(
        default_factory=lambda: MessageConfig(source="", content=""),
        sa_column=Column(JSON),
    )
    session_id: Optional[int] = Field(
        default=None,
        sa_column=Column(Integer, ForeignKey("session.id", ondelete="CASCADE")),
    )
    run_id: Optional[int] = Field(
        default=None,
        sa_column=Column(Integer, ForeignKey("run.id", ondelete="CASCADE")),
    )
    message_meta: Optional[Union[MessageMeta, dict[str, Any]]] = Field(
        default={}, sa_column=Column(JSON)
    )


class Session(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    user_id: Optional[str] = None
    version: Optional[str] = "0.0.1"
    team_id: Optional[int] = Field(
        default=None,
        sa_column=Column(Integer, ForeignKey("team.id", ondelete="CASCADE")),
    )
    name: Optional[str] = None
    agent: Optional[str] = Field(default="magentic-one")


class RunStatus(str, Enum):
    CREATED = "created"
    ACTIVE = "active"
    COMPLETE = "complete"
    ERROR = "error"
    STOPPED = "stopped"
    AWAITING_INPUT = "awaiting_input"
    PAUSED = "paused"


class InputType(str, Enum):
    TEXT_INPUT = "text_input"
    APPROVAL = "approval"


class Run(SQLModel, table=True):
    """Represents a single execution run within a session"""

    __table_args__ = {"sqlite_autoincrement": True}

    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )
    session_id: Optional[int] = Field(
        default=None,
        sa_column=Column(
            Integer, ForeignKey("session.id", ondelete="CASCADE"), nullable=False
        ),
    )
    status: RunStatus = Field(default=RunStatus.CREATED)

    # Store the original user task
    task: Union[MessageConfig, dict[str, Any]] = Field(
        default_factory=lambda: MessageConfig(source="", content=""),
        sa_column=Column(JSON),
    )

    # Store TeamResult which contains TaskResult
    team_result: Union[TeamResult, dict[str, Any]] = Field(
        default=None, sa_column=Column(JSON)
    )

    error_message: Optional[str] = None
    version: Optional[str] = "0.0.1"
    messages: Union[List[Message], List[dict[str, Any]]] = Field(
        default_factory=list, sa_column=Column(JSON)
    )

    user_id: Optional[str] = None
    state: Optional[str] = None

    input_request: Optional[dict[str, Any]] = Field(
        default=None, sa_column=Column(JSON)
    )

    @field_serializer("created_at", "updated_at")
    def serialize_datetime(cls, value: datetime) -> str:
        if isinstance(value, datetime):
            return value.isoformat()


class Gallery(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    user_id: Optional[str] = None
    version: Optional[str] = "0.0.1"
    config: Union[GalleryConfig, dict[str, Any]] = Field(
        default_factory=lambda: GalleryConfig(
            id="",
            name="",
            metadata=GalleryMetadata(author="", version=""),
            components=GalleryComponents(
                agents=[], models=[], tools=[], terminations=[], teams=[]
            ),
        ),
        sa_column=Column(JSON),
    )

    @field_serializer("created_at", "updated_at")
    def serialize_datetime(cls, value: datetime) -> str:
        if isinstance(value, datetime):
            return value.isoformat()


class Settings(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    user_id: Optional[str] = None
    version: Optional[str] = "0.0.1"
    config: Union[SettingsConfig, dict[str, Any]] = Field(
        default_factory=SettingsConfig, sa_column=Column(JSON)
    )


class Plan(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    user_id: Optional[str] = None
    version: Optional[str] = "0.0.1"
    task: Optional[str] = None
    steps: Optional[List[dict[str, Any]]] = Field(
        default_factory=list, sa_column=Column(JSON)
    )
    session_id: Optional[int] = None

    @field_serializer("created_at", "updated_at")
    def serialize_datetime(cls, value: datetime) -> str:
        if isinstance(value, datetime):
            return value.isoformat()


## 更新的部分

### agent mode setting
class AgentModeSettings(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    user_id: Optional[str] = None
    version: Optional[str] = "0.0.1"
    config: Union[AgentModeSetting, dict[str, Any]] = Field(
        default_factory=AgentModeSetting, sa_column=Column(JSON)
    )

class AgentModeConfig(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    user_id: Optional[str] = None
    version: Optional[str] = "0.0.1"
    mode: Optional[str] = "drsai"
    config: Optional[dict[str, Any]] = Field(
        default_factory=dict, sa_column=Column(JSON)
    )

### user files

class UserFiles(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    user_id: Optional[str] = None
    session_id: Optional[int] = None
    version: Optional[str] = "0.0.1"
    files: Optional[dict[str, Any]] = Field(
        default_factory=dict, sa_column=Column(JSON)
    )


### user remote agent

class UserAgents(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    user_id: Optional[str] = None
    version: Optional[str] = "0.0.1"
    agents: Optional[dict[str, Any]] = Field(
        default_factory=dict, sa_column=Column(JSON)
    )

##

DatabaseModel = Team | Message | Session | Run | Gallery | Settings | Plan | AgentModeSettings | AgentModeConfig | UserAgents



# api/deps.py
import logging
from contextlib import contextmanager
from typing import Any, Dict, Optional
from pathlib import Path
from fastapi import HTTPException, status

from ..database import DatabaseManager
from .config import settings
from .managers.connection import WebSocketManager

logger = logging.getLogger(__name__)

# Global manager instances
_db_manager: Optional[DatabaseManager] = None
_websocket_manager: Optional[WebSocketManager] = None

# Context manager for database sessions


@contextmanager
def get_db_context():
    """Provide a transactional scope around a series of operations."""
    if not _db_manager:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Database manager not initialized",
        )
    try:
        yield _db_manager
    except Exception as e:
        logger.error(f"Database operation failed: {str(e)}")
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Database operation failed",
        ) from e


# Dependency providers


async def get_db() -> DatabaseManager:
    """Dependency provider for database manager"""
    if not _db_manager:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Database manager not initialized",
        )
    return _db_manager


async def get_websocket_manager() -> WebSocketManager:
    """Dependency provider for connection manager"""
    if not _websocket_manager:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Connection manager not initialized",
        )
    return _websocket_manager


# Manager initialization and cleanup


async def init_managers(
    database_uri: str,
    config_dir: Path,
    app_root: Path,
    internal_workspace_root: str,
    external_workspace_root: str,
    inside_docker: bool,
    config: Dict[str, Any],
) -> None:
    """Initialize all manager instances"""
    global _db_manager, _websocket_manager, _team_manager

    logger.info("Initializing managers...")

    try:
        # Initialize database manager
        _db_manager = DatabaseManager(engine_uri=database_uri, base_dir=app_root)
        _db_manager.initialize_database(auto_upgrade=settings.UPGRADE_DATABASE)

        # init default team config
        await _db_manager.import_teams_from_directory(
            config_dir, settings.DEFAULT_USER_ID, check_exists=True
        )

        # Initialize connection manager
        _websocket_manager = WebSocketManager(
            db_manager=_db_manager,
            internal_workspace_root=Path(internal_workspace_root),
            external_workspace_root=Path(external_workspace_root),
            inside_docker=inside_docker,
            config=config,
        )
        logger.info("Connection manager initialized")

    except Exception as e:
        logger.error(f"Failed to initialize managers: {str(e)}")
        await cleanup_managers()  # Cleanup any partially initialized managers
        raise


async def cleanup_managers() -> None:
    """Cleanup and shutdown all manager instances"""
    global _db_manager, _websocket_manager, _team_manager

    logger.info("Cleaning up managers...")

    # Cleanup connection manager first to ensure all active connections are closed
    if _websocket_manager:
        try:
            await _websocket_manager.cleanup()
        except Exception as e:
            logger.error(f"Error cleaning up connection manager: {str(e)}")
        finally:
            _websocket_manager = None

    # TeamManager doesn't need explicit cleanup since WebSocketManager handles it
    _team_manager = None

    # Cleanup database manager last
    if _db_manager:
        try:
            await _db_manager.close()
        except Exception as e:
            logger.error(f"Error cleaning up database manager: {str(e)}")
        finally:
            _db_manager = None

    logger.info("All managers cleaned up")


# Utility functions for dependency management


# Error handling for manager operations


class ManagerOperationError(Exception):
    """Custom exception for manager operation errors"""

    def __init__(self, manager_name: str, operation: str, detail: str):
        self.manager_name = manager_name
        self.operation = operation
        self.detail = detail
        super().__init__(f"{manager_name} failed during {operation}: {detail}")



import os
import logging
import requests
import pandas as pd
from typing import List, Union
from ...benchmark import Benchmark
from ...models import BaseTask, BaseCandidate, BaseEvalResult


class WebGamesBenchmark(Benchmark):
    """
    Loads the WebGames dataset from Hugging Face and evaluates predictions
    by comparing against the known passwords.
    """

    DATASET_REPO_ID = "convergence-ai/webgames"
    TEST_FILE = "test.jsonl"

    def __init__(
        self,
        name: str = "WebGames",
        data_dir: Union[str, None] = None,
        base_website_path: str = "https://webgames.convergence.ai/",
    ):
        """
        Benchmark from https://github.com/convergence-ai/webgames

        It is best to host WebGames yourself and set the base_website_path to the url of the hosted website.
        To host WebGames:
            ```bash
            git clone https://github.com/convergence-ai/webgames.git
            cd webgames
            npm install
            npm run dev
            ```
            You will find it http://localhost:5173 and you can pass base_website_path to the benchmark to use your own hosted version.

        Args:
            base_website_path: The base path of the website to use for the WebGamesBenchmark. Make sure it ends with a slash.
        """
        assert data_dir is not None, "data_dir must be provided for WebGamesBenchmark"
        super().__init__(name=name, data_dir=data_dir)
        self.eval_result_class = BaseEvalResult
        self.base_website_path = base_website_path
        logging_msg: str = (
            f"[WebGames] Using base website path: {self.base_website_path}"
        )
        if self.base_website_path == "https://webgames.convergence.ai/":
            logging_msg += """
            You can also run the benchmark with a hosted version of the website by setting the base_website_path to the url of the hosted website.
            To host WebGames:
            ```bash
            git clone https://github.com/convergence-ai/webgames.git
            cd webgames
            npm install
            npm run dev
            ```
            You will find it http://localhost:5173 and you can pass base_website_path to the benchmark to use your own hosted version.
        
            """
        logging.info(logging_msg)

    def download_dataset(self) -> None:
        """
        Download the dataset into self.data_dir using huggingface datasets.
        """
        assert self.data_dir is not None
        if not os.path.exists(self.data_dir):
            os.makedirs(self.data_dir, exist_ok=True)

        logging.info(f"[WebGames] Downloading dataset into '{self.data_dir}'...")

        # Use pandas to download and save the dataset
        df = pd.read_json(  # type: ignore
            f"hf://datasets/{self.DATASET_REPO_ID}/{self.TEST_FILE}", lines=True
        )
        output_path = os.path.join(self.data_dir, self.TEST_FILE)
        df.to_json(output_path, orient="records", lines=True)  # type: ignore

        logging.info("[WebGames] Dataset downloaded.")

    def load_dataset(self) -> None:
        """
        Read in the test.jsonl file and store tasks.
        """
        # Double check that the base website path is valid and is reachable
        try:
            response = requests.get(self.base_website_path)
            response.raise_for_status()
        except Exception as e:
            raise ValueError(
                f"Invalid base website path: {self.base_website_path}"
            ) from e

        assert self.data_dir is not None
        test_path = os.path.join(self.data_dir, self.TEST_FILE)

        if not os.path.isfile(test_path):
            raise FileNotFoundError(
                f"Could not find {self.TEST_FILE} in {self.data_dir}. "
                "Make sure you have downloaded the dataset."
            )

        # Load test set using pandas
        df = pd.read_json(test_path, lines=True)  # type: ignore
        added_instruction: str = "There are no errors in the website. You need to complete the task on this website and follow the instruction untill a password is revealed. A password will only be revealed if you complete the task correctly. Do not navigate away from this website."

        for _, row in df.iterrows():
            task = BaseTask(
                id=row["id"],  # type: ignore
                question=f"{added_instruction}\n\n{row['description']}",  # type: ignore
                ground_truth=row["password"],  # type: ignore
                url_path=f"{self.base_website_path}{row['path']}",  # type: ignore
                metadata={"title": row["title"], "tags": row["tags"]},  # type: ignore
                set="test",
            )
            self.tasks[task.id] = task

        logging.info(f"[WebGames] Loaded {len(self.tasks)} total examples.")

    def get_split_tasks(self, split: str) -> List[str]:
        """
        Returns task IDs for the specified split (only 'test' is available).
        """
        if split != "test":
            raise ValueError("only 'test' split is available for WebGames")
        return [task_id for task_id, task in self.tasks.items() if task.set == split]

    def evaluator(self, task: BaseTask, candidate: BaseCandidate) -> BaseEvalResult:
        """
        Evaluate if the candidate password matches the ground truth password.
        """
        # Cast to proper types if needed
        if isinstance(task, dict):
            task = BaseTask(**task)  # type: ignore
        if isinstance(candidate, dict):
            candidate = BaseCandidate(**candidate)  # type: ignore
        # check if the ground truth password is anywhere in the candidate answer, as a substring
        score = 1.0 if task.ground_truth in candidate.answer else 0.0

        return BaseEvalResult(score=score)



import base64
import json
from typing import AsyncGenerator, Sequence
from autogen_agentchat.base import Response
from autogen_agentchat.messages import BaseChatMessage, TextMessage
from autogen_core import CancellationToken
from autogen_agentchat.messages import (
    MultiModalMessage,
)
from openai import OpenAI


from ._web_surfer import WebSurfer

# This code has been adapted from https://github.com/openai/openai-cua-sample-app
# Copyright 2025 OpenAI - MIT License


class WebSurferCUA(WebSurfer):
    """A dummy version of WebSurfer that returns static responses for testing and CUA purposes."""

    async def on_messages_stream(
        self, messages: Sequence[BaseChatMessage], cancellation_token: CancellationToken
    ) -> AsyncGenerator[BaseChatMessage | Response, None]:
        """Dummy implementation that returns a static response."""
        # add the last message to the chat history
        if not hasattr(self, "_cua_chat_history"):
            self._cua_chat_history = []
        if isinstance(messages[-1], TextMessage):
            self._cua_chat_history.append(
                {
                    "role": "user",
                    "content": messages[-1].content,
                }
            )
        elif isinstance(messages[-1], MultiModalMessage):
            self._cua_chat_history.append(
                {
                    "role": "user",
                    "content": messages[-1].content,
                }
            )

        client = OpenAI()

        new_items = []
        assert self._page is not None
        actions_so_far = 0
        while (
            actions_so_far < self.max_actions_per_step
            and new_items[-1].get("role") != "assistant"
            if new_items
            else True
        ):
            response = client.responses.create(
                model="computer-use-preview",
                tools=[
                    {
                        "type": "computer_use_preview",
                        "display_width": self._page.viewport_size["width"],
                        "display_height": self._page.viewport_size["height"],
                        "environment": "browser",
                    },
                    {
                        "type": "function",
                        "name": "back",
                        "description": "Go back to the previous page.",
                        "parameters": {},
                    },
                    {
                        "type": "function",
                        "name": "goto",
                        "description": "Go to a specific URL.",
                        "parameters": {
                            "type": "object",
                            "properties": {
                                "url": {
                                    "type": "string",
                                    "description": "Fully qualified URL to navigate to.",
                                },
                            },
                            "additionalProperties": False,
                            "required": ["url"],
                        },
                    },
                    {
                        "type": "function",
                        "name": "create_tab",
                        "description": "Creates a new browser tab and navigates to the specified URL.",
                        "parameters": {
                            "type": "object",
                            "properties": {
                                "url": {
                                    "type": "string",
                                    "description": "The URL to open in the new tab.",
                                },
                            },
                            "required": ["url"],
                        },
                    },
                ],
                input=self._cua_chat_history + new_items,
                reasoning={
                    "generate_summary": "concise",
                },
                truncation="auto",
            )
            response = response.to_dict()
            if "output" not in response:
                raise ValueError("No output from CUA")
            else:
                new_items += response["output"]
                for item in response["output"]:
                    output_item = []
                    if item["type"] == "function_call":
                        name, args = item["name"], json.loads(item["arguments"])

                        yield Response(
                            chat_message=TextMessage(
                                content=f"{name}({args})",
                                source=self.name,
                            )
                        )
                        output_item = await self.handle_function_call(item)
                    if item["type"] == "computer_call":
                        action = item["action"]
                        action_type = action["type"]
                        action_args = {k: v for k, v in action.items() if k != "type"}
                        yield Response(
                            chat_message=TextMessage(
                                content=f"{action_type}({action_args})",
                                source=self.name,
                            )
                        )
                        output_item = await self.handle_computer_call(item)
                    if item["type"] == "message":
                        output_item = []
                        yield Response(
                            chat_message=TextMessage(
                                content=item["content"][0]["text"],
                                source=self.name,
                            )
                        )
                    if output_item is not None and output_item != []:
                        new_items += output_item
            actions_so_far += 1
        self._cua_chat_history += new_items

    async def handle_function_call(self, item):
        name, args = item["name"], json.loads(item["arguments"])
        if True:
            print(f"{name}({args})")

        if name == "back":
            await self._playwright_controller.go_back(self._page)
        elif name == "goto":
            await self._playwright_controller.visit_page(self._page, url=args["url"])
        elif name == "create_tab":
            new_page = await self._playwright_controller.create_new_tab(
                self._context, url=args["url"]
            )
            self._page = new_page
        return [
            {
                "type": "function_call_output",
                "call_id": item["call_id"],
                "output": "success",
            }
        ]

    async def handle_computer_call(self, item):
        # native CUA actions
        assert self._page is not None  # Ensure page is not None
        action = item["action"]
        action_type = action["type"]
        action_args = {k: v for k, v in action.items() if k != "type"}
        if True:
            print(f"{action_type}({action_args})")

        if True:  # Keep print for debugging if desired
            print(f"Executing CUA action: {action_type}({action_args})")

        # Map CUA actions to PlaywrightController methods
        if action_type == "click":
            # Assuming 'x', 'y', and optional 'button' are in action_args
            await self._playwright_controller.click_coords(
                self._page,
                x=action_args["x"],
                y=action_args["y"],
                button=action_args.get("button", "left"),
            )
        elif action_type == "double_click":
            # Assuming 'x', 'y' are in action_args
            await self._playwright_controller.double_click_coords(
                self._page, x=action_args["x"], y=action_args["y"]
            )
        elif action_type == "scroll":
            # Assuming 'x', 'y', 'scroll_x', 'scroll_y' are in action_args
            # Note: Playwright's scroll_coords requires start coords (x, y)
            await self._playwright_controller.scroll_coords(
                self._page,
                x=action_args.get("x", 0),  # Default start coords if not provided
                y=action_args.get("y", 0),
                scroll_x=action_args["scroll_x"],
                scroll_y=action_args["scroll_y"],
            )
        elif action_type == "wait":
            # Assuming 'duration' (in seconds) is in action_args
            await self._playwright_controller.sleep(
                self._page, duration=action_args.get("ms", 1000) / 1000
            )
        elif action_type == "move":
            # Assuming 'x', 'y' are in action_args (maps to hover)
            await self._playwright_controller.hover_coords(
                self._page, x=action_args["x"], y=action_args["y"]
            )
        elif action_type == "keypress":
            # Assuming 'keys' (list of strings) is in action_args
            await self._playwright_controller.keypress(
                self._page, keys=action_args["keys"]
            )
        elif action_type == "drag":
            # Assuming 'path' (list of dicts with 'x', 'y') is in action_args
            await self._playwright_controller.drag_coords(
                self._page, path=action_args["path"]
            )
        elif action_type == "type":
            await self._playwright_controller.type_direct(
                self._page, text=action_args["text"]
            )

        screenshot = await self._playwright_controller.get_screenshot(self._page)
        screenshot_base64 = base64.b64encode(screenshot).decode("utf-8")
        # if user doesn't ack all safety checks exit with error
        pending_checks = item.get("pending_safety_checks", [])
        for check in pending_checks:
            message = check["message"]
            if False:  # Replace False with actual check logic if needed
                raise ValueError(
                    f"Safety check failed: {message}. Cannot continue with unacknowledged safety checks."
                )

        call_output = {
            "type": "computer_call_output",
            "call_id": item["call_id"],
            "acknowledged_safety_checks": pending_checks,
            # Default to success if no specific output (like screenshot)
            "output": {
                "type": "input_image",
                "image_url": f"data:image/png;base64,{screenshot_base64}",
            },
        }

        current_url, _ = await self._playwright_controller.get_current_url_title(
            self._page
        )
        call_output["output"]["current_url"] = current_url

        return [call_output]



# api/app.py
import os
import yaml
from dotenv import load_dotenv
load_dotenv()
from contextlib import asynccontextmanager
from typing import AsyncGenerator, Any

# import logging
from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from loguru import logger

from ...version import VERSION
from .config import settings
from .deps import cleanup_managers, init_managers
from .initialization import AppInitializer
from .routes import (
    plans,
    runs,
    sessions,
    settingsroute,
    teams,
    validation,
    ws,
    agent_mode,
    files,
    agent_worker,
)
import httpx
from fastapi.responses import HTMLResponse
# Initialize application
app_file_path = os.path.dirname(os.path.abspath(__file__))
initializer = AppInitializer(settings, app_file_path)


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
    """
    Lifecycle manager for the FastAPI application.
    Handles initialization and cleanup of application resources.
    """

    try:
        # Load the config if provided
        config: dict[str, Any] = {}
        config_file = os.environ.get("_CONFIG")
        if config_file:
            with open(config_file, "r") as f:
                config = yaml.safe_load(f)

        # Initialize managers (DB, Connection, Team)
        await init_managers(
            initializer.database_uri,
            initializer.config_dir,
            initializer.app_root,
            os.environ["INTERNAL_WORKSPACE_ROOT"],
            os.environ["EXTERNAL_WORKSPACE_ROOT"],
            os.environ["INSIDE_DOCKER"] == "1",
            config,
        )

        # Any other initialization code
        logger.info(
            f"Application startup complete. Navigate to http://{os.environ.get('_HOST', '127.0.0.1')}:{os.environ.get('_PORT', '8081')}"
        )

    except Exception as e:
        logger.error(f"Failed to initialize application: {str(e)}")
        raise

    yield  # Application runs here

    # Shutdown
    try:
        logger.info("Cleaning up application resources...")
        await cleanup_managers()
        logger.info("Application shutdown complete")
    except Exception as e:
        logger.error(f"Error during shutdown: {str(e)}")


# Create FastAPI application
app = FastAPI(lifespan=lifespan, debug=True)

# CORS middleware configuration
app.add_middleware(
    CORSMiddleware,
    allow_origins=[
        "http://localhost:8000",
        "http://127.0.0.1:8000",
        "http://localhost:8001",
        "http://localhost:8081",
        "http://drsai.ihep.ac.cn",
        "https://drsai.ihep.ac.cn",
        "https://aitest.ihep.ac.cn",
    ],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# Create API router with version and documentation
api = FastAPI(
    root_path="/api",
    title="DrSai-UI API",
    version=VERSION,
    description="DrSai-UI API is an application to interact with web agents.",
    docs_url="/docs" if settings.API_DOCS else None,
)

# Include all routers with their prefixes
api.include_router(
    sessions.router,
    prefix="/sessions",
    tags=["sessions"],
    responses={404: {"description": "Not found"}},
)

api.include_router(
    plans.router,
    prefix="/plans",
    tags=["plans"],
    responses={404: {"description": "Not found"}},
)

api.include_router(
    runs.router,
    prefix="/runs",
    tags=["runs"],
    responses={404: {"description": "Not found"}},
)

api.include_router(
    teams.router,
    prefix="/teams",
    tags=["teams"],
    responses={404: {"description": "Not found"}},
)


api.include_router(
    ws.router,
    prefix="/ws",
    tags=["websocket"],
    responses={404: {"description": "Not found"}},
)

api.include_router(
    validation.router,
    prefix="/validate",
    tags=["validation"],
    responses={404: {"description": "Not found"}},
)

api.include_router(
    settingsroute.router,
    prefix="/settings",
    tags=["settings"],
    responses={404: {"description": "Not found"}},
)

# 添加的新路由

api.include_router(
    agent_mode.router,
    prefix="/agentmode",
    tags=["agentmode"],
    responses={404: {"description": "Not found"}},
)

files.initializer=initializer
api.include_router(
    files.router,
    prefix="/files",
    tags=["files"],
    responses={404: {"description": "Not found"}},
)

api.include_router(
    agent_worker.router,
    prefix="/agentworker",
    tags=["agentworker"],
    responses={404: {"description": "Not found"}},
)

# Version endpoint


@api.get("/version")
async def get_version():
    """Get API version"""
    return {
        "status": True,
        "message": "Version retrieved successfully",
        "data": {"version": VERSION},
    }


# Health check endpoint


@api.get("/health")
async def health_check():
    """API health check endpoint"""
    return {
        "status": True,
        "message": "Service is healthy",
    }

# 加载vnc api
from .vnc_router import router as vnc_router
api.include_router(vnc_router, prefix="/vncapi", tags=["vnc"])

# Mount static file directories
app.mount("/api", api)


# 加载统一认证模块
SERVICE_MODE = os.getenv("SERVICE_MODE", None)
if SERVICE_MODE == "PROD":
    from ....drsai_adapter.sso.ihep_sso_router import router as ihep_sso_router
    from ....drsai_adapter.sso.ihep_sso_router import oauth_config
    from starlette.middleware.sessions import SessionMiddleware
    app.add_middleware(SessionMiddleware, secret_key=oauth_config.meddleware_secret)
    app.include_router(ihep_sso_router, prefix="/umt", tags=["umt"])


app.mount(
    "/files",
    StaticFiles(directory=initializer.static_root, html=True),
    name="files",
)
app.mount("/", StaticFiles(directory=initializer.ui_root, html=True), name="ui")

# Error handlers



@app.exception_handler(500)
async def internal_error_handler(request: Request, exc: Exception):
    logger.error(f"Internal error: {str(exc)}")
    return {
        "status": False,
        "message": "Internal server error",
        "detail": str(exc) if settings.API_DOCS else "Internal server error",
    }


def create_app() -> FastAPI:
    """
    Factory function to create and configure the FastAPI application.
    Useful for testing and different deployment scenarios.
    """
    return app



import io
from typing import BinaryIO, Dict, List, Tuple, cast

from PIL import Image, ImageDraw, ImageFont

from ...tools.playwright.types import DOMRectangle, InteractiveRegion

"""
This module provides functionality to annotate screenshots with numbered markers for interactive regions.
It handles marking visible elements as well as tracking elements that are above or below the viewport.
"""

TOP_NO_LABEL_ZONE = 20  # Don't print any labels close the top of the page


def add_set_of_mark(
    screenshot: bytes | Image.Image | io.BufferedIOBase,
    ROIs: Dict[str, InteractiveRegion],
    use_sequential_ids: bool = False,
) -> Tuple[Image.Image, List[str], List[str], List[str], Dict[str, str]]:
    """
    Add numbered markers to a screenshot for each interactive region.

    Args:
        screenshot (bytes | Image.Image | io.BufferedIOBase): The screenshot image as bytes, PIL Image, or file-like object
        ROIs (Dict[str, InteractiveRegion]): Dictionary mapping element IDs to their interactive regions
        use_sequential_ids (bool): If True, assigns sequential numbers to elements instead of using original IDs

    Returns:
        Tuple containing:
        - Image.Image: Annotated image
        - List[str]: List of visible element IDs
        - List[str]: List of element IDs above viewport
        - List[str]: List of element IDs below viewport
        - Dict[str, str]: Mapping of displayed IDs to original element IDs
    """
    if isinstance(screenshot, Image.Image):
        return _add_set_of_mark(screenshot, ROIs, use_sequential_ids)

    if isinstance(screenshot, bytes):
        screenshot = io.BytesIO(screenshot)

    image = Image.open(cast(BinaryIO, screenshot))
    comp, visible_rects, rects_above, rects_below, id_mapping = _add_set_of_mark(
        image, ROIs, use_sequential_ids
    )
    image.close()
    return comp, visible_rects, rects_above, rects_below, id_mapping


def _add_set_of_mark(
    screenshot: Image.Image,
    ROIs: Dict[str, InteractiveRegion],
    use_sequential_ids: bool = True,
) -> Tuple[Image.Image, List[str], List[str], List[str], Dict[str, str]]:
    """
    Internal implementation for adding markers to the screenshot.

    Args:
        screenshot (Image.Image): PIL Image to annotate
        ROIs (Dict[str, InteractiveRegion]): Dictionary of interactive regions
        use_sequential_ids (bool): Whether to use sequential numbers instead of original IDs

    Returns:
        Same as :func:`add_set_of_mark`
    """
    visible_rects: List[str] = []
    rects_above: List[str] = []  # Scroll up to see
    rects_below: List[str] = []  # Scroll down to see
    id_mapping: Dict[str, str] = {}  # Maps new IDs to original IDs

    base = screenshot.convert("RGBA")

    # First pass to identify all rectangles
    for original_id, roi in ROIs.items():
        # Handle options separately and add to visible only
        if roi.get("tag_name") == "option" or roi.get("tag_name") == "input, type=file":
            if original_id not in visible_rects:
                visible_rects.append(original_id)
            continue

        # Check each rectangle for the element
        for rect in roi["rects"]:
            if not rect or rect["width"] * rect["height"] == 0:
                continue

            mid = (
                (rect["right"] + rect["left"]) / 2.0,
                (rect["top"] + rect["bottom"]) / 2.0,
            )

            # Only process if x coordinate is valid
            if 0 <= mid[0] < base.size[0]:
                # Add to exactly one list based on y coordinate
                if mid[1] < 0 and original_id not in rects_above:
                    rects_above.append(original_id)
                elif mid[1] >= base.size[1] and original_id not in rects_below:
                    rects_below.append(original_id)
                elif 0 <= mid[1] < base.size[1] and original_id not in visible_rects:
                    visible_rects.append(original_id)

    # Create new sequential IDs for all rectangles
    next_id = 1
    original_to_new: Dict[str, str] = {}  # Add reverse mapping

    # Helper function to create new mapped IDs
    def map_ids(original_ids: List[str]) -> List[str]:
        nonlocal next_id, original_to_new
        new_ids: List[str] = []
        for original_id in original_ids:
            new_id = str(next_id)
            id_mapping[new_id] = original_id
            original_to_new[original_id] = new_id  # Store reverse mapping
            new_ids.append(new_id)
            next_id += 1
        return new_ids

    if use_sequential_ids:
        # Map IDs in sequence: visible first, then above, then below
        new_visible_rects = map_ids(visible_rects)
        new_rects_above = map_ids(rects_above)
        new_rects_below = map_ids(rects_below)
    else:
        # Use original IDs but still maintain the mapping
        new_visible_rects = visible_rects.copy()
        new_rects_above = rects_above.copy()
        new_rects_below = rects_below.copy()
        # Create identity mapping for all IDs
        for id_list in [visible_rects, rects_above, rects_below]:
            for original_id in id_list:
                id_mapping[original_id] = original_id
                original_to_new[original_id] = original_id

    # Drawing code remains the same
    fnt = ImageFont.load_default(14)
    overlay = Image.new("RGBA", base.size)
    draw = ImageDraw.Draw(overlay)

    # Much simpler and faster lookup using the reverse mapping
    for original_id, roi in ROIs.items():
        if roi.get("tag_name") == "option":
            continue

        new_id = original_to_new.get(original_id)
        if new_id is None:
            continue  # Skip if no mapping found

        for rect in roi["rects"]:
            if not rect or rect["width"] * rect["height"] == 0:
                continue

            mid = (
                (rect["right"] + rect["left"]) / 2.0,
                (rect["top"] + rect["bottom"]) / 2.0,
            )

            if 0 <= mid[0] < base.size[0] and 0 <= mid[1] < base.size[1]:
                _draw_roi(draw, new_id, fnt, rect)

    comp = Image.alpha_composite(base, overlay)
    overlay.close()

    return comp, new_visible_rects, new_rects_above, new_rects_below, id_mapping


def _draw_roi(
    draw: ImageDraw.ImageDraw,
    idx: str | int,  # Fix type hint to allow both string and int indices
    font: ImageFont.FreeTypeFont | ImageFont.ImageFont,
    rect: DOMRectangle,
) -> None:
    """
    Draw a single region of interest on the image.

    Args:
        draw (ImageDraw.ImageDraw): PIL ImageDraw object
        idx (str | int): Index/ID to display on the marker
        font (ImageFont.FreeTypeFont | ImageFont.ImageFont): Font to use for the marker text
        rect (DOMRectangle): Rectangle coordinates for the region
    """
    color = (255, 0, 0, 255)  # Red color for outline
    text_color = (255, 255, 255, 255)  # White text for better contrast

    roi = ((rect["left"], rect["top"]), (rect["right"], rect["bottom"]))

    # Adjust label position if too close to top of screen
    label_location = (rect["right"], rect["top"])
    label_anchor = "rb"

    if label_location[1] <= TOP_NO_LABEL_ZONE:
        label_location = (rect["right"], rect["bottom"])
        label_anchor = "rt"

    draw.rectangle(roi, outline=color, width=2)

    bbox = draw.textbbox(
        label_location, str(idx), font=font, anchor=label_anchor, align="center"
    )
    bbox = (bbox[0] - 3, bbox[1] - 3, bbox[2] + 3, bbox[3] + 3)
    draw.rectangle(bbox, fill=color)

    draw.text(
        label_location,
        str(idx),
        fill=text_color,
        font=font,
        anchor=label_anchor,
        align="center",
    )



import os
import warnings
import typer
import uvicorn
from typing_extensions import Annotated
from typing import Optional
from pathlib import Path
import logging

from ..version import VERSION
from ..version import APP_NAME as UI_APP_NAME
from .._docker import (
    check_docker_running,
    check_browser_image,
    check_python_image,
    build_browser_image,
    build_python_image,
)

from .web.app import app as ui_app

logging.basicConfig(level=logging.ERROR)

app = typer.Typer()

# Ignore deprecation warnings from websockets
warnings.filterwarnings("ignore", message="websockets.legacy is deprecated*")
warnings.filterwarnings(
    "ignore", message="websockets.server.WebSocketServerProtocol is deprecated*"
)


def get_env_file_path():
    app_dir = os.path.join(os.path.expanduser("~"), ".drsai_ui")
    if not os.path.exists(app_dir):
        os.makedirs(app_dir, exist_ok=True)
    return os.path.join(app_dir, "temp_env_vars.env")


@app.command()
def ui(
    host: str = "127.0.0.1",
    port: int = 8081,
    workers: int = 1,
    reload: Annotated[bool, typer.Option("--reload")] = False,
    docs: bool = True,
    appdir: str = str(Path.home() / ".drsai_ui"),
    database_uri: Optional[str] = None,
    upgrade_database: bool = False,
    config: Optional[str] = None,
    rebuild_docker: Optional[bool] = False,
):
    """
    Run Magentic-UI.

    Args:
        host (str, optional): Host to run the UI on. Defaults to 127.0.0.1 (localhost).
        port (int, optional): Port to run the UI on. Defaults to 8081.
        workers (int, optional): Number of workers to run the UI with. Defaults to 1.
        reload (bool, optional): Whether to reload the UI on code changes. Defaults to False.
        docs (bool, optional): Whether to generate API docs. Defaults to False.
        appdir (str, optional): Path to the app directory where files are stored. Defaults to None.
        database-uri (str, optional): Database URI to connect to. Defaults to None.
        config (str, optional): Path to the config file. Defaults to `config.yaml`.
        rebuild_docker: bool, optional: Rebuild the docker images. Defaults to False.
    """

    typer.echo(typer.style(f"Starting {UI_APP_NAME}", fg=typer.colors.GREEN, bold=True))

    # Set things up for Docker
    typer.echo("Checking if Docker is running...", nl=False)

    if not check_docker_running():
        typer.echo(typer.style("Failed\n", fg=typer.colors.RED, bold=True))
        typer.echo("Docker is not running. Please start Docker and try again.")
        raise typer.Exit(1)
    else:
        typer.echo(typer.style("OK", fg=typer.colors.GREEN, bold=True))

    typer.echo("Checking Docker vnc browser image...", nl=False)
    if not check_browser_image() or rebuild_docker:
        typer.echo(typer.style("Update\n", fg=typer.colors.YELLOW, bold=True))
        typer.echo("Building Docker vnc image (this WILL take a few minutes)")
        build_browser_image()
        typer.echo("\n")
    else:
        typer.echo(typer.style("OK", fg=typer.colors.GREEN, bold=True))

    typer.echo("Checking Docker python image...", nl=False)
    if not check_python_image() or rebuild_docker:
        typer.echo(typer.style("Update\n", fg=typer.colors.YELLOW, bold=True))
        typer.echo("Building Docker python image (this WILL take a few minutes)")
        build_python_image()
        typer.echo("\n")
    else:
        typer.echo(typer.style("OK", fg=typer.colors.GREEN, bold=True))

    # check the images again and throw an error if they are not found
    if not check_browser_image() or not check_python_image():
        typer.echo(typer.style("Failed\n", fg=typer.colors.RED, bold=True))
        typer.echo("Docker images not found. Please build the images and try again.")
        raise typer.Exit(1)

    typer.echo("Launching Web Application...")

    # Write configuration
    env_vars = {
        "_HOST": host,
        "_PORT": port,
        "_API_DOCS": str(docs),
    }

    if appdir:
        env_vars["_APPDIR"] = appdir
    if database_uri:
        env_vars["DATABASE_URI"] = database_uri
    if upgrade_database:
        env_vars["_UPGRADE_DATABASE"] = "1"

    env_vars["INSIDE_DOCKER"] = "0"
    env_vars["EXTERNAL_WORKSPACE_ROOT"] = appdir
    env_vars["INTERNAL_WORKSPACE_ROOT"] = appdir

    # If the config file is not provided, check for the default config file
    if not config:
        if os.path.isfile("config.yaml"):
            config = "config.yaml"
        else:
            typer.echo("Config file not provided. Using default settings.")
    if config:
        env_vars["_CONFIG"] = config

    # Create temporary env file to share configuration with uvicorn workers
    env_file_path = get_env_file_path()
    with open(env_file_path, "w") as temp_env:
        for key, value in env_vars.items():
            temp_env.write(f"{key}={value}\n")

    uvicorn.run(
        # "drsai.backend.ui.ui_backend.backend.web.app:app",
        ui_app,
        host=host,
        port=port,
        workers=workers,
        reload=reload,
        reload_excludes=["**/alembic/*", "**/alembic.ini", "**/versions/*"]
        if reload
        else None,
        env_file=env_file_path,
    )


@app.command()
def version():
    """
    Print the version of the Magentic-UI backend CLI.
    """

    typer.echo(f"Magentic-UI version: {VERSION}")


def run():
    app()


if __name__ == "__main__":
    app()



from pathlib import Path
from typing import Tuple
import socket
from autogen_core import ComponentModel
from .vnc_docker_playwright_browser import VncDockerPlaywrightBrowser

import random

from dotenv import load_dotenv
load_dotenv()
import os

def get_available_port(min_port: int = int(os.getenv("MIN_PORT", 40000)), max_port: int = int(os.getenv("MAX_PORT", 65000))) -> tuple[int, socket.socket]:
    """
    Get an available port on the local machine.
    """
    # s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    # s.bind(("127.0.0.1", 0))
    # port = s.getsockname()[1]
    # return port, s

    max_attempts = 100  # 最大随机尝试次数
    for _ in range(max_attempts):
        port = random.randint(min_port, max_port)
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.bind(("127.0.0.1", port))
            return port, s
        except OSError:
            s.close()

    # 如果随机尝试失败，改为顺序扫描
    for port in range(min_port, max_port):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.bind(("127.0.0.1", port))
            return port, s
        except OSError:
            s.close()
    
    raise OSError(f"No available ports in range {min_port}-{max_port}")


def get_browser_resource_config(
    bind_dir: Path,
    novnc_port: int = -1,
    playwright_port: int = -1,
    inside_docker: bool = True,
) -> Tuple[ComponentModel, int, int]:
    """
    Create a VNC Docker Playwright Browser Resource configuration. The requested ports for novnc and playwright may be overwritten. The final values for each port number will be in the return value.

    Args:
        bind_dir (str): Directory to bind for the browser resource.
        novnc_port (int, optional): Port for the noVNC server. Default: -1 (auto-assign).
        playwright_port (int, optional): Port for the Playwright browser. Default: -1 (auto-assign).

    Returns:
        A tuple containing the following:
            - VncDockerPlaywrightBrowserResource: Configured browser resource.
            - int: Port number for the noVNC server.
            - int: Port number for the Playwright browser.
    """
    opened_sockets: list[socket.socket] = []

    if novnc_port == -1:
        novnc_port, sock = get_available_port()
        opened_sockets.append(sock)
    if playwright_port == -1:
        playwright_port, sock = get_available_port()
        opened_sockets.append(sock)

    # Close the sockets after getting the ports
    for sock in opened_sockets:
        sock.close()

    return (
        VncDockerPlaywrightBrowser(
            bind_dir=bind_dir,
            playwright_port=playwright_port,
            novnc_port=novnc_port,
            inside_docker=inside_docker,
        ).dump_component(),
        novnc_port,
        playwright_port,
    )



from .playwright_controller import PlaywrightController
from .playwright_state import BrowserState
from .types import (
    InteractiveRegion,
    VisualViewport,
    domrectangle_from_dict,
)
from .browser import (
    PlaywrightBrowser,
    DockerPlaywrightBrowser,
    LocalPlaywrightBrowser,
    VncDockerPlaywrightBrowser,
    HeadlessDockerPlaywrightBrowser,
)

__all__ = [
    "PlaywrightController",
    "BrowserState",
    "InteractiveRegion",
    "VisualViewport",
    "domrectangle_from_dict",
    "PlaywrightBrowser",
    "DockerPlaywrightBrowser",
    "LocalPlaywrightBrowser",
    "VncDockerPlaywrightBrowser",
    "HeadlessDockerPlaywrightBrowser",
]



import asyncio

from typing import Union, Any, Dict, List, Optional, Callable
from autogen_core.models import ChatCompletionClient, LLMMessage
from autogen_core import ComponentModel, CancellationToken

from autogen_core.tools._function_tool import FunctionTool, FunctionToolConfig
# from autogen_agentchat.agents._assistant_agent import AssistantAgentConfig
# from autogen_agentchat.agents import UserProxyAgent, BaseChatAgent
from autogen_agentchat.teams import BaseGroupChat
from autogen_ext.tools.mcp import (
    SseServerParams, 
    StdioServerParams, 
    mcp_server_tools,
    SseMcpToolAdapter,
    StdioMcpToolAdapter,
    )
from autogen_core.tools._function_tool import FunctionTool, FunctionToolConfig
from autogen_core.code_executor._func_with_reqs import (
    Import, 
    import_to_str, 
    to_code,
    ImportFromModule, 
    Alias
    )
from drsai import (
    AssistantAgent, 
    DrSaiRoundRobinGroupChat, 
    DrSaiSelectorGroupChat, 
    DrSaiSwarm)
from .magentic_one.agents.drsai_agents import MagenticAgent
from .magentic_one.teams.orchestrator import GroupChat as MagenticGroupChat
from .components.memory.load_memory_cofig import load_memory_function

def get_model_client(
        model_client_config: Union[ComponentModel, Dict[str, Any], None],
    ) -> ChatCompletionClient:
        return ChatCompletionClient.load_component(model_client_config)

async def a_load_mcp_tools(mcp_tools_config: list[dict]) -> list[StdioMcpToolAdapter | SseMcpToolAdapter]:
    '''
    加载AutoGen的MCP工具
    yaml:
        mcp_tools_1: &tool_1
            type: std
            command: python3
            args:
                - /your/path/to/mcp_server.py
            env: null

            mcp_tools_2: &tool_2
            type: sse
            url: www.example.com
            headers:
                Authorization: Bearer <token>
                event_name: mcp_tools_event
            timeout: 20
            sse_read_timeout: 300
    '''
    mcp_tools = []
    for mcp_tool_config in mcp_tools_config:

        if mcp_tool_config["type"] == "std":
            command = mcp_tool_config.get("command", None)
            assert command is not None, "mcp_tools的command不能为空"
            args = mcp_tool_config.get("args", [])
            env = mcp_tool_config.get("env", None)
            cwd = mcp_tool_config.get("cwd", None)
            std_tool: list[StdioMcpToolAdapter] = await mcp_server_tools(StdioServerParams(
                                command=command,
                                args=args,
                                env=env,
                                cwd=cwd)
                        ) 
            mcp_tools.extend(std_tool)
        elif mcp_tool_config["type"] == "sse":
            url = mcp_tool_config.get("url", None)
            assert url is not None, "mcp_tools的url不能为空"
            headers = mcp_tool_config.get("headers", None)
            timeout = mcp_tool_config.get("timeout", 20)
            sse_read_timeout = mcp_tool_config.get("sse_read_timeout", 60 * 5)
            sse_tool: list[SseMcpToolAdapter] = await mcp_server_tools(SseServerParams(
                                url=url,
                                headers=headers,
                                timeout=timeout,
                                sse_read_timeout=sse_read_timeout)
                        ) 
            mcp_tools.extend(sse_tool)
        else:
            raise ValueError(f"mcp_tools的type只能是std或sse")
    return mcp_tools

async def a_load_hepai_tools(hepai_tools_config: list[dict]) -> list[Callable]:
    '''
    加载HepAI的worker工具
    yaml:
        hepai_tools_1: &hepai_tool_1
            worker_name: besiii_boss_08
            api_key: sk-****
            allowed_tools:
                - json_mapping
                - job_status
            base_url: "https://aiapi.ihep.ac.cn/apiv2"
    '''
    hepai_tools = []
    try:
        from hepai.tools.get_woker_functions import get_worker_async_functions
        for hepai_tool_config in hepai_tools_config:
            worker_name = hepai_tool_config.get("worker_name", None)
            api_key = hepai_tool_config.get("api_key", None)
            base_url = hepai_tool_config.get("base_url", "https://aiapi.ihep.ac.cn/apiv2")
            funcs: list[Callable] = await get_worker_async_functions(worker_name, api_key, base_url)
            allowed_tools = hepai_tool_config.get("allowed_tools", [])
            if allowed_tools:
                funcs = [func for func in funcs if func.__name__ in allowed_tools]
            hepai_tools.extend(funcs)
        return hepai_tools
    except ImportError:
        raise ImportError("please install <hepai> package")

async def a_load_source_code_tools(source_tools_config: list[dict]) -> list[FunctionTool]:
    '''
    加载前端源代码工具
    yaml:
        source_tools_1: &source_tool_1
            type: FunctionTool
            config:
                source_code: |
                    def calculator(a: float, b: float, operator: str) -> str:
                        try:
                            if operator == "+":
                                return str(a + b)
                            elif operator == "-":
                                return str(a - b)
                            elif operator == "*":
                                return str(a * b)
                            elif operator == "/":
                                if b == 0:
                                    return "Error: Division by zero"
                                return str(a / b)
                            else:
                                return "Error: Invalid operator. Please use +, -, *, or /"
                        except Exception as e:
                            return f"Error: {str(e)}"
                name: calculator
                description: A simple calculator that performs basic arithmetic operations
                global_imports:
                    - math: 
                        - pi
                        - e
                has_cancellation_support: False
    '''
    source_code_tools = []
    for source_tool_config in source_tools_config:
        if source_tool_config["type"] == "FunctionTool":
            global_imports = source_tool_config["config"].get("global_imports", [])
            global_imports_modules = []
            if global_imports:
                for global_import in global_imports:
                    module_name = list(global_import.keys())[0]
                    import_list = global_import[module_name]
                    global_imports_modules.append(ImportFromModule(module=module_name, imports=import_list))
                source_tool_config["config"]["global_imports"] = global_imports_modules
            tool_config = FunctionToolConfig(**source_tool_config["config"])
            tool = FunctionTool._from_config(tool_config)
            source_code_tools.append(tool)
        else:
            raise ValueError(f"source_tools的type只能是FunctionTool")
    return source_code_tools

async def a_load_agent_factory_from_config(
        config: dict,
        mode = "backend"
        ) -> Callable[[], Union[AssistantAgent, BaseGroupChat]]:
    '''
    加载配置，创建AssistantAgent或BaseGroupChat实例
    '''
    
    assistant_list = []
    groupchat: BaseGroupChat|None = None
    for key, value in config.items():
        if isinstance(value, dict) and "type" in value.keys():
             
            if value["type"] == "AssistantAgent":
                name = value.get("name", "AssistantAgent")
                system_message = value.get("system_message", None)
                description = value.get("description", None)
                model_client = get_model_client(value.get("model_client", None))
                # 加载tools
                mcp_tools = await a_load_mcp_tools(value.get("mcp_tools", []))
                hepai_tools = await a_load_hepai_tools(value.get("hepai_tools", []))
                source_tools = await a_load_source_code_tools(value.get("source_tools", []))
                # TODO: 加载 agent memory_functions
                # TODO: 加载RAG等knowledge function
                knowledge_cofig = value.get("knowledge_cofig", None)
                memory_function = None
                if knowledge_cofig:
                    memory_function = await load_memory_function(knowledge_cofig)
                assistant_list.append(AssistantAgent(
                    name=name,
                    system_message=system_message,
                    description=description,
                    model_client=model_client,
                    tools=mcp_tools+hepai_tools if len(mcp_tools+hepai_tools+source_tools) > 0 else None,
                    memory_function=memory_function,
                ))
                
    # TODO: 完善GroupChat的加载, UI模式下的Groupchat需要MagenticGroupChat

    # TODO: 通过本地PIP安装的智能体/多智能体系统进行通用加载，需要实现a_load_agent_factory_from_config

    # TODO: 通过远程启动智能体/多智能体系统进行通用加载-》基于MagenticAgent实现一个智能体OpenAPI接口，可以暂停/关闭等远程模型连接的功能
    assert len(assistant_list) > 0, "AssistantAgent配置不能为空"
    async def agent_factory() -> AssistantAgent:
        return assistant_list[0]
    return agent_factory

async def a_load_agent_factory_from_mode(
        config: dict,
        mode = "backend"
        ) -> Callable[[], Union[AssistantAgent, BaseGroupChat]]:
    '''
    加载配置，创建AssistantAgent或BaseGroupChat实例
    '''
    name = config.get("name", "AssistantAgent")
    system_message = config.get("system_message", None)
    description = config.get("description", None)
    model_client = get_model_client(config.get("model_client", None))
    # 加载tools
    mcp_tools = await a_load_mcp_tools(config.get("mcp_tools", []))
    hepai_tools = await a_load_hepai_tools(config.get("hepai_tools", []))
    source_tools = await a_load_source_code_tools(config.get("source_tools", []))
    # TODO: 加载 agent memory_functions
    # TODO: 加载RAG等knowledge function
    knowledge_cofig = config.get("knowledge_cofig", None)
    memory_function = None
    if knowledge_cofig:
        memory_function = await load_memory_function(knowledge_cofig)
    if mode == "ui":
        agent = MagenticAgent(
            name=name,
            system_message=system_message,
            description=description,
            model_client=model_client,
            tools=mcp_tools+hepai_tools+source_tools if len(mcp_tools+hepai_tools+source_tools) > 0 else None,
            memory_function=memory_function,
        )
    else:
        agent = AssistantAgent( 
            name=name,
            system_message=system_message,
            description=description,
            model_client=model_client,
            tools=mcp_tools+hepai_tools if len(mcp_tools+hepai_tools) > 0 else None,
            memory_function=memory_function,
        )
    async def agent_factory() -> AssistantAgent:
        return agent
    return agent_factory






from .input_func import AsyncInputFunc, InputFuncType, InputRequestType, SyncInputFunc
from .version import __version__

ABOUT = "Magentic-UI - A web browsing assistant."
__all__ = [
    "AsyncInputFunc",
    "InputFuncType",
    "InputRequestType",
    "SyncInputFunc",
    "__version__",
]
